import logging
from typing import Optional, List, Any, ClassVar, Dict

import requests  # type: ignore
from pydantic import Field

from bearish.exceptions import InvalidApiKeyError
from bearish.exchanges.exchanges import Countries
from bearish.models.assets.assets import Assets
from bearish.models.assets.equity import Equity
from bearish.models.assets.etfs import Etf
from bearish.models.base import SourceBase
from bearish.models.financials.balance_sheet import BalanceSheet
from bearish.models.financials.base import Financials
from bearish.models.financials.cash_flow import CashFlow
from bearish.models.financials.metrics import FinancialMetrics
from bearish.models.price.price import Price
from bearish.models.query.query import AssetQuery
from bearish.sources.base import AbstractSource, ApiUsage
from bearish.types import Sources, SeriesLength
from bearish.utils.utils import get_start_date

logger = logging.getLogger(__name__)

API_URL = "https://financialmodelingprep.com/api/v3/"


def compose_url(  # noqa: PLR0913
    api_url: str,
    endpoint: str,
    api_key: str,
    ticker: str,
    period: Optional[str] = None,
    from_: Optional[str] = None,
) -> str:
    period_ = f"?period={period}" if period else ""
    from__ = f"?from={from_}" if from_ else ""
    if from_ and period:
        raise NotImplementedError("Cannot use both period and from_")
    separator = "&" if (period or from_) else "?"
    return f"{api_url}{endpoint}/{ticker}{period_}{from__}{separator}apikey={api_key}"


def read_api(  # noqa: PLR0913
    api_url: str,
    endpoint: str,
    api_key: str,
    ticker: str,
    period: Optional[str] = None,
    from_: Optional[str] = None,
) -> Any:
    request_response = requests.get(
        compose_url(api_url, endpoint, api_key, ticker, period=period, from_=from_),
        timeout=10,
    )
    response_json = request_response.json()
    if isinstance(response_json, dict) and response_json.get("Error Message"):
        logger.error(f"Error reading {ticker}: {response_json['Error Message']}")
        raise InvalidApiKeyError(response_json["Error Message"])
    if isinstance(response_json, list):
        for response_json_ in response_json:
            if response_json_.get("Error Message"):
                logger.error(
                    f"Error reading {ticker}: {response_json_['Error Message']}"
                )
                raise InvalidApiKeyError(response_json_["Error Message"])
    return response_json


class FmpSourceBase(SourceBase):
    __source__: Sources = "FMP"


class FmpAssetsSourceBase(SourceBase):
    __source__: Sources = "FMPAssets"


class FmpAssetsEquity(FmpAssetsSourceBase, Equity):
    __alias__: ClassVar[Dict[str, str]] = {
        "symbol": "symbol",
        "name": "name",
        "exchange": "exchange",
        "exchangeShortName": "market",
    }


class FmpAssetsEtf(FmpAssetsSourceBase, Etf):
    __alias__: ClassVar[Dict[str, str]] = {
        "symbol": "symbol",
        "name": "name",
        "exchange": "exchange",
        "exchangeShortName": "market",
    }


class FmpEquity(FmpSourceBase, Equity):
    __alias__: ClassVar[Dict[str, str]] = {
        "symbol": "symbol",  # Ticker symbol
        "lastDiv": "dividend_rate",  # Last dividend rate
        "companyName": "name",  # Company name
        "currency": "currency",  # Currency
        "isin": "isin",  # ISIN
        "cusip": "cusip",  # CUSIP
        "exchange": "exchange",  # Exchange name
        "exchangeShortName": "market",  # Abbreviated market name
        "industry": "industry",  # Industry
        "website": "website",  # Company website
        "description": "summary",  # Company summary
        "sector": "sector",  # Sector
        "country": "country",  # Country
        "city": "city",  # City
        "state": "state",  # State
        "zip": "zipcode",  # ZIP code
        "name": "name",  # Company name
        "marketCap": "market_capitalization",  # Market capitalization
        "eps": "trailing_earnings_per_share",  # Trailing EPS
        "pe": "trailing_price_to_earnings",  # Trailing P/E ratio
        "sharesOutstanding": "shares_outstanding",  # Outstanding shares
        "dividendYielTTM": "dividend_yield",  # TTM dividend yield
        "peRatioTTM": "trailing_price_to_earnings",  # Trailing P/E ratio
        "pegRatioTTM": "trailing_peg_ratio",  # Trailing PEG ratio
        "currentRatioTTM": "current_ratio",  # Current ratio
        "quickRatioTTM": "quick_ratio",  # Quick ratio
        "grossProfitMarginTTM": "gross_margins",  # Gross margins
        "operatingProfitMarginTTM": "operating_margins",  # Operating margins
        "returnOnAssetsTTM": "return_on_assets",  # ROA
        "returnOnEquityTTM": "return_on_equity",  # ROE
    }

    @classmethod
    def from_tickers(cls, tickers: List[str]) -> List["FmpEquity"]:
        tickers_ = []
        for ticker in tickers:
            try:
                profile = read_api(API_URL, "profile", cls.__api_key__, ticker)
                quote = read_api(API_URL, "quote", cls.__api_key__, ticker)
                ratio_ttm = read_api(API_URL, "ratios-ttm", cls.__api_key__, ticker)
                key_metrics = read_api(
                    API_URL, "key-metrics-ttm", cls.__api_key__, ticker
                )
                datas = [*profile, *quote, *ratio_ttm, *key_metrics]
                data = {k: v for data in datas for k, v in data.items()}
                tickers_.append(cls.model_validate(data))
            except InvalidApiKeyError as e:  # noqa: PERF203
                logger.warning(f"Error reading {ticker}: {e}")
                break
            except Exception as e:
                logger.error(f"Error reading {ticker}: {e}")
                continue
        return tickers_


class FmpFinancialMetrics(FmpSourceBase, FinancialMetrics):
    __alias__: ClassVar[Dict[str, str]] = {
        "peRatioTTM": "pe_ratio",
        "marketCapTTM": "market_capitalization",
        "netProfitMarginTTM": "profit_margin",
        "netIncomePerShareTTM": "basic_eps",
        "revenuePerShareTTM": "total_revenue",
        "grossProfitMarginTTM": "gross_profit",
    }


class FmpBalanceSheet(FmpSourceBase, BalanceSheet):
    __alias__: ClassVar[Dict[str, str]] = {
        "date": "date",
        "symbol": "symbol",
        "cashAndCashEquivalents": "cash_and_cash_equivalents_at_carrying_value",
        "cashAndShortTermInvestments": "cash_and_short_term_investments",
        "netReceivables": "current_net_receivables",
        "inventory": "inventory",
        "otherCurrentAssets": "other_current_assets",
        "totalCurrentAssets": "total_current_assets",
        "propertyPlantEquipmentNet": "property_plant_equipment",
        "totalAssets": "total_assets",
        "totalCurrentLiabilities": "total_current_liabilities",
        "otherCurrentLiabilities": "other_current_liabilities",
        "shortTermDebt": "current_debt",
        "accountPayables": "current_accounts_payable",
        "totalNonCurrentLiabilities": "total_non_current_liabilities",
        "otherNonCurrentLiabilities": "other_non_current_liabilities",
        "longTermDebt": "long_term_debt",
        "totalLiabilities": "total_liabilities",
        "retainedEarnings": "retained_earnings",
        "commonStock": "common_stock",
        "totalStockholdersEquity": "total_shareholder_equity",
        "capitalLeaseObligations": "capital_lease_obligations",
    }


class FmpCashFlow(FmpSourceBase, CashFlow):
    __alias__: ClassVar[Dict[str, str]] = {
        "date": "date",
        "symbol": "symbol",
        "netIncome": "net_income_from_continuing_operations",
        "depreciationAndAmortization": "depreciation_amortization_depletion",
        "changeInWorkingCapital": "change_in_working_capital",
        "accountsReceivables": "change_in_receivables",
        "inventory": "change_in_inventory",
        "accountsPayables": "change_in_operating_liabilities",
        "otherWorkingCapital": "change_in_other_working_capital",
        "netCashProvidedByOperatingActivities": "operating_cash_flow",
        "investmentsInPropertyPlantAndEquipment": "capital_expenditure",
        "netCashUsedForInvestingActivites": "cash_flow_from_investing_activities",
        "commonStockIssued": "proceeds_from_issuance_of_common_stock",
        "commonStockRepurchased": "repurchase_of_capital_stock",
        "dividendsPaid": "cash_dividends_paid",
        "netCashUsedProvidedByFinancingActivities": "financing_cash_flow",
        "netChangeInCash": "changes_in_cash",
    }


class FmpPrice(FmpSourceBase, Price):
    __alias__: ClassVar[Dict[str, str]] = {
        "date": "date",
        "open": "open",
        "high": "high",
        "low": "low",
        "close": "close",
        "volume": "volume",
    }


class FmpSource(FmpSourceBase, AbstractSource):
    countries: List[Countries] = ["US"]  # noqa: RUF012
    api_usage: ApiUsage = ApiUsage(calls_limit=220)

    def set_api_key(self, api_key: str) -> None:
        FmpSourceBase.__api_key__ = api_key

    def _read_assets(self, query: Optional[AssetQuery] = None) -> Assets:
        if query is None:
            return Assets()

        if query.symbols.empty():
            return Assets()
        equities = FmpEquity.from_tickers(query.symbols.equities_symbols())
        return Assets(equities=equities)

    def _read_financials(self, tickers: List[str]) -> List[Financials]:
        financials = []
        for ticker in tickers:
            balance_sheet_statement = read_api(
                API_URL,
                "balance-sheet-statement",
                self.__api_key__,
                ticker,
                period="annual",
            )
            cash_flow_statement = read_api(
                API_URL,
                "cash-flow-statement",
                self.__api_key__,
                ticker,
                period="annual",
            )
            ratio_ttm = read_api(API_URL, "ratios-ttm", self.__api_key__, ticker)
            key_metrics = read_api(API_URL, "key-metrics-ttm", self.__api_key__, ticker)
            self.api_usage.add_api_calls(4)
            datas = [*ratio_ttm, *key_metrics]
            financial_metrics = {k: v for data in datas for k, v in data.items()}
            financial_metrics.update({"symbol": ticker})
            financials.append(
                Financials(
                    financial_metrics=[
                        FmpFinancialMetrics.model_validate(financial_metrics)
                    ],
                    balance_sheets=[
                        FmpBalanceSheet.model_validate(balance_sheet_statement_)
                        for balance_sheet_statement_ in balance_sheet_statement
                    ],
                    cash_flows=[
                        FmpCashFlow.model_validate(cash_flow_statement_)
                        for cash_flow_statement_ in cash_flow_statement
                    ],
                )
            )
        return financials

    def _read_series(self, tickers: List[str], type: SeriesLength) -> List[FmpPrice]:  # type: ignore
        from_ = get_start_date(type)
        prices = []
        for ticker in tickers:
            historical_price = read_api(
                API_URL,
                "historical-price-full",
                self.__api_key__,
                ticker,
                period=None,
                from_=from_,
            )
            self.api_usage.add_api_calls(1)
            symbol = historical_price["symbol"]
            datas = historical_price["historical"]
            prices.extend(
                [FmpPrice.model_validate({**data, "symbol": symbol}) for data in datas]
            )
        return prices


class FmpAssetsSource(FmpAssetsSourceBase, AbstractSource):
    countries: List[Countries] = Field(default_factory=list)

    def _read_assets(self, query: Optional[AssetQuery] = None) -> Assets:
        stocks = read_api(API_URL, "stock", self.__api_key__, "list")
        stocks = [s for s in stocks if s["type"] == "stock"]
        etfs = read_api(API_URL, "etf", self.__api_key__, "list")
        etfs = [s for s in etfs if s["type"] == "etf"]
        return Assets(
            equities=[FmpAssetsEquity.model_validate(stock) for stock in stocks],
            etfs=[FmpAssetsEtf.model_validate(etf) for etf in etfs],
        )

    def _read_financials(self, tickers: List[str]) -> List[Financials]:
        return [Financials()]

    def _read_series(self, tickers: List[str], type: SeriesLength) -> List[Price]:
        return []

    def set_api_key(self, api_key: str) -> None:
        FmpAssetsSourceBase.__api_key__ = api_key



from typing import Annotated, Optional

from pydantic import BeforeValidator, Field

from bearish.models.base import DataSourceBase
from bearish.utils.utils import to_float


class CashFlow(DataSourceBase):
    operating_cash_flow: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(
            default=None, description="Cash flow generated from operating activities"
        ),
    ]
    change_in_operating_liabilities: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(default=None, description="Changes in operating liabilities"),
    ]
    change_in_working_capital: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(
            default=None,
            description="Changes in working capital from operating activities",
        ),
    ]
    change_in_other_working_capital: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(
            default=None,
            description="Changes in other working capital from operating activities",
        ),
    ]
    change_in_receivables: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(default=None, description="Changes in accounts receivable balance"),
    ]
    change_in_inventory: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(default=None, description="Changes in inventory levels"),
    ]
    depreciation_amortization_depletion: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(
            default=None,
            description="Total depreciation, depletion, and amortization expenses",
        ),
    ]
    capital_expenditure: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(default=None, description="Capital expenditures for fixed assets"),
    ]
    cash_flow_from_investing_activities: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(
            default=None,
            description="Cash flow generated or used in investing activities",
        ),
    ]
    financing_cash_flow: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(
            default=None,
            description="Cash flow generated or used in financing activities",
        ),
    ]
    repurchase_of_capital_stock: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(default=None, description="Payments made for repurchase of common stock"),
    ]
    cash_dividends_paid: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(default=None, description="Cash dividends paid to shareholders"),
    ]
    common_stock_dividend_paid: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(default=None, description="Dividends paid on common stock"),
    ]
    proceeds_from_issuance_of_common_stock: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(default=None, description="Proceeds from issuance of common stock"),
    ]
    changes_in_cash: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(default=None, description="Net changes in cash and cash equivalents"),
    ]
    net_income_from_continuing_operations: Annotated[
        Optional[float],
        BeforeValidator(to_float),
        Field(default=None, description="Net income from continuing operations"),
    ]


class QuarterlyCashFlow(CashFlow): ...



import datetime
import logging
import os
from enum import Enum
from pathlib import Path
from typing import Optional, List, Any, get_args, Annotated, cast, Union, Type, Callable

import typer
from pydantic import (
    BaseModel,
    Field,
    ConfigDict,
    PrivateAttr,
    validate_call,
    model_validator,
)
from rich.console import Console


from bearish.database.crud import BearishDb
from bearish.exceptions import InvalidApiKeyError, LimitApiKeyReachedError
from bearish.exchanges.exchanges import (
    Countries,
    exchanges_factory,
    ExchangeQuery,
    Exchanges,
)
from bearish.interface.interface import BearishDbBase
from bearish.models.api_keys.api_keys import SourceApiKeys
from bearish.models.assets.assets import Assets
from bearish.models.base import Ticker, TrackerQuery, FinancialsTracker, PriceTracker
from bearish.models.financials.base import Financials
from bearish.models.price.price import Price
from bearish.models.price.prices import Prices
from bearish.models.query.query import AssetQuery, Symbols
from bearish.sources.base import AbstractSource
from bearish.sources.financedatabase import FinanceDatabaseSource
from bearish.sources.financial_modelling_prep import FmpAssetsSource, FmpSource
from bearish.sources.investpy import InvestPySource
from bearish.sources.tiingo import TiingoSource
from bearish.sources.yahooquery import YahooQuerySource
from bearish.sources.yfinance import yFinanceSource
from bearish.types import SeriesLength, Sources
from bearish.utils.utils import batch

logger = logging.getLogger(__name__)
app = typer.Typer()
console = Console()


class CountryEnum(str, Enum): ...


CountriesEnum = Enum(  # type: ignore
    "CountriesEnum",
    {country: country for country in get_args(Countries)},
    type=CountryEnum,
)


class Filter(BaseModel):
    countries: List[CountriesEnum]
    filters: Optional[List[str] | str] = None

    @model_validator(mode="after")
    def _model_validator(self) -> "Filter":
        if self.filters is not None and isinstance(self.filters, str):
            self.filters = [t.strip() for t in self.filters.split(",")]
        return self

    def filter(self, tickers: List[Ticker]) -> List[Ticker]:
        if not self.filters:
            return tickers
        return list({t for t in tickers if t.symbol in self.filters})


class Bearish(BaseModel):
    model_config = ConfigDict(extra="forbid")
    path: Path
    auto_migration: bool = True
    batch_size: int = Field(default=500)
    api_keys: SourceApiKeys = Field(default_factory=SourceApiKeys)
    _bearish_db: BearishDbBase = PrivateAttr()
    exchanges: Exchanges = Field(default_factory=exchanges_factory)
    asset_sources: List[AbstractSource] = Field(
        default_factory=lambda: [
            FinanceDatabaseSource(),
            InvestPySource(),
            FmpAssetsSource(),
        ]
    )
    detailed_asset_sources: List[AbstractSource] = Field(
        default_factory=lambda: [yFinanceSource(), YahooQuerySource(), FmpSource()]
    )
    financials_sources: List[AbstractSource] = Field(
        default_factory=lambda: [
            yFinanceSource(),
            YahooQuerySource(),
            FmpSource(),
        ]
    )
    price_sources: List[AbstractSource] = Field(
        default_factory=lambda: [
            yFinanceSource(),
            YahooQuerySource(),
            TiingoSource(),
        ]
    )

    def model_post_init(self, __context: Any) -> None:
        self._bearish_db = BearishDb(
            database_path=self.path, auto_migration=self.auto_migration
        )
        for source in set(
            self.financials_sources
            + self.price_sources
            + self.asset_sources
            + self.detailed_asset_sources
        ):
            try:
                source.set_api_key(
                    self.api_keys.keys.get(
                        source.__source__, os.environ.get(source.__source__.upper())  # type: ignore
                    )
                )
            except Exception as e:  # noqa: PERF203
                logger.error(
                    f"Invalid API key for {source.__source__}: {e}. It will be removed from sources"
                )
                for sources in [
                    self.financials_sources,
                    self.price_sources,
                    self.asset_sources,
                    self.detailed_asset_sources,
                ]:
                    if source in sources:
                        sources.remove(source)

    def get_asset_sources(self) -> List[Sources]:
        return [source.__source__ for source in self.asset_sources]

    def get_detailed_asset_sources(self) -> List[Sources]:
        return [source.__source__ for source in self.detailed_asset_sources]

    def write_assets(self, query: Optional[AssetQuery] = None) -> None:
        existing_sources = self._bearish_db.read_sources()
        asset_sources = [
            asset_source
            for asset_source in self.asset_sources
            if asset_source.__source__ not in existing_sources
        ]
        logger.debug(f"Found asset sources: {[s.__source__ for s in asset_sources]}")
        return self._write_base_assets(asset_sources, query)

    def write_detailed_assets(self, query: Optional[AssetQuery] = None) -> None:
        return self._write_base_assets(
            self.detailed_asset_sources, query, use_all_sources=False
        )

    def _write_base_assets(
        self,
        asset_sources: List[AbstractSource],
        query: Optional[AssetQuery] = None,
        use_all_sources: bool = True,
    ) -> None:

        if query:
            cached_assets = self.read_assets(AssetQuery.model_validate(query))
            query.update_symbols(cached_assets)
        for source in asset_sources:

            logger.info(f"Fetching assets from source {type(source).__name__}")
            assets_ = source.read_assets(query)
            if assets_.is_empty():
                logger.warning(f"No assets found from {type(source).__name__}")
                continue
            logger.debug(
                f"writing assets from {type(source).__name__}. Number of symbols: {len(assets_.symbols())}"
            )
            self._bearish_db.write_assets(assets_)
            self._bearish_db.write_source(source.__source__)
            if use_all_sources:
                continue
            if not assets_.failed_query.symbols:
                break
            else:
                query = AssetQuery(
                    symbols=Symbols(equities=assets_.failed_query.symbols)  # type: ignore
                )

    def read_assets(self, assets_query: AssetQuery) -> Assets:
        return self._bearish_db.read_assets(assets_query)

    def read_financials(self, assets_query: AssetQuery) -> Financials:
        return self._bearish_db.read_financials(assets_query)

    def read_series(self, assets_query: AssetQuery, months: int = 1) -> List[Price]:
        return self._bearish_db.read_series(assets_query, months=months)

    def _get_tracked_tickers(
        self,
        tracker_query: TrackerQuery,
        tracker_type: Union[Type[PriceTracker], Type[FinancialsTracker]],
    ) -> List[Ticker]:
        return self._bearish_db.read_tracker(tracker_query, tracker_type)

    def get_tickers_without_financials(self, tickers: List[Ticker]) -> List[Ticker]:
        tracked_tickers = self._get_tracked_tickers(TrackerQuery(), FinancialsTracker)
        return [t for t in tickers if t not in tracked_tickers]

    def get_tickers_without_price(self, tickers: List[Ticker]) -> List[Ticker]:
        return [
            t
            for t in tickers
            if t not in self._get_tracked_tickers(TrackerQuery(), PriceTracker)
        ]

    def get_ticker_with_price(self) -> List[Ticker]:
        return [
            Ticker(symbol=t)
            for t in self._get_tracked_tickers(TrackerQuery(), PriceTracker)
        ]

    def write_many_financials(self, tickers: List[Ticker]) -> None:
        logger.warning(f"Found tickers without financials: {len(tickers)}")
        chunks = batch(tickers, size=self.batch_size)
        source = self.financials_sources[0]

        for chunk in chunks:
            logger.debug(f"getting financial data for {len(chunk)} tickers")
            try:
                financials_ = source.read_financials(chunk)
            except (InvalidApiKeyError, LimitApiKeyReachedError, Exception) as e:
                logger.error(f"Error reading data using {source.__source__}: {e}")
                continue

            if not financials_:
                logger.warning("No financial data found.")
                continue
            self._bearish_db.write_financials(financials_)
            self._bearish_db.write_trackers(
                [
                    FinancialsTracker(
                        symbol=t.symbol,
                        source=source.__source__,
                        exchange=t.exchange,
                    )
                    for t in chunk
                ]
            )

    @validate_call
    def write_many_series(self, tickers: List[Ticker], type: SeriesLength) -> None:
        chunks = batch(tickers, self.batch_size)
        source = self.price_sources[0]
        for chunk in chunks:
            logger.debug(f"getting financial data for {len(chunk)} tickers")
            try:
                series_ = source.read_series(chunk, type)
            except (InvalidApiKeyError, LimitApiKeyReachedError, Exception) as e:
                logger.error(f"Error reading series: {e}")
                continue
            if series_:
                price_date = Prices(prices=series_).get_last_date()
                self._bearish_db.write_series(series_)
                self._bearish_db.write_trackers(
                    [
                        PriceTracker(
                            symbol=t.symbol,
                            source=source.__source__,
                            exchange=t.exchange,
                            date=price_date,
                        )
                        for t in chunk
                    ]
                )

    def read_sources(self) -> List[str]:
        return self._bearish_db.read_sources()

    def get_tickers(self, exchange_query: ExchangeQuery) -> List[Ticker]:
        return self._bearish_db.get_tickers(exchange_query)

    def get_detailed_tickers(self, filter: Filter) -> None:
        tickers = self.get_tickers(
            self.exchanges.get_exchange_query(
                cast(List[Countries], filter.countries), self.get_asset_sources()
            )
        )

        tickers = filter.filter(tickers)
        asset_query = AssetQuery(symbols=Symbols(equities=tickers))  # type: ignore
        self.write_detailed_assets(asset_query)

    def get_financials(self, filter: Filter) -> None:
        tickers = self.get_tickers(
            self.exchanges.get_exchange_query(
                cast(List[Countries], filter.countries),
                self.get_detailed_asset_sources(),
            )
        )
        tickers = filter.filter(tickers)
        logger.debug(f"Found tickers: {[t.symbol for t in tickers]}")
        self.write_many_financials(tickers)

    def get_prices(self, filter: Filter) -> None:
        tickers = self.get_tickers(
            self.exchanges.get_exchange_query(
                cast(List[Countries], filter.countries),
                self.get_detailed_asset_sources(),
            )
        )
        tickers = filter.filter(tickers)
        self.write_many_series(tickers, "max")

    def _update(
        self,
        tracker_type: Union[Type[PriceTracker], Type[FinancialsTracker]],
        write_function: Callable[[List[Ticker]], None],
        symbols: Optional[List[str]] = None,
        reference_date: Optional[datetime.date] = None,
        delay: int = 5,
    ) -> None:
        reference_date = reference_date or datetime.date.today()
        tickers = self._get_tracked_tickers(
            TrackerQuery(reference_date=reference_date, delay=delay), tracker_type
        )
        if symbols is not None:
            tickers = [t for t in tickers if t.symbol in symbols]
        write_function(tickers)

    def update_prices(
        self,
        symbols: Optional[List[str]] = None,
        reference_date: Optional[datetime.date] = None,
        delay: int = 1,
        series_length: SeriesLength = "1mo",
    ) -> None:
        def write_function(tickers: List[Ticker]) -> None:
            logger.debug(f"Updating prices for {len(tickers)} tickers")
            self.write_many_series(tickers, series_length)

        self._update(
            PriceTracker,
            write_function,
            symbols=symbols,
            reference_date=reference_date,
            delay=delay,
        )

    def update_financials(
        self,
        symbols: Optional[List[str]] = None,
        reference_date: Optional[datetime.date] = None,
        delay: int = 20,
    ) -> None:
        def write_function(tickers: List[Ticker]) -> None:
            self.write_many_financials(tickers)

        self._update(
            FinancialsTracker,
            write_function,
            symbols=symbols,
            reference_date=reference_date,
            delay=delay,
        )


@app.command()
def run(
    path: Path,
    countries: Annotated[List[CountriesEnum], typer.Argument()],
    filters: Optional[str] = None,
    api_keys: Optional[Path] = None,
) -> None:

    console.log(
        f"Fetching assets to database for countries: {countries}, with filters: {filters}",
    )
    source_api_keys = SourceApiKeys.from_file(api_keys)
    bearish = Bearish(path=path, api_keys=source_api_keys)
    with console.status("[bold green]Fetching Tickers data..."):
        bearish.write_assets()
        filter = Filter(countries=countries, filters=filters)
        bearish.get_detailed_tickers(filter)
        console.log("[bold][red]Tickers downloaded!")
    with console.status("[bold green]Fetching Financial data..."):
        bearish.get_financials(filter)
        console.log("[bold][red]Financial downloaded!")
    with console.status("[bold green]Fetching Price data..."):
        bearish.get_prices(filter)
        console.log("[bold][red]Price downloaded!")


@app.command()
def tickers(
    path: Path,
    countries: Annotated[List[CountriesEnum], typer.Argument()],
    filters: Optional[str] = None,
    api_keys: Optional[Path] = None,
    skip_base_tickers: Optional[bool] = False,
) -> None:
    with console.status("[bold green]Fetching Tickers data..."):
        logger.info(
            f"Writing assets to database for countries: {countries}",
        )
        source_api_keys = SourceApiKeys.from_file(api_keys)
        bearish = Bearish(path=path, api_keys=source_api_keys)
        if not skip_base_tickers:
            console.log("[green]Fetching base Tickers[/green]")
            bearish.write_assets()
        filter = Filter(countries=countries, filters=filters)
        console.log("[green]Fetching detailed Tickers[/green]")
        bearish.get_detailed_tickers(filter)
        console.log("[bold][red]Tickers downloaded!")


@app.command()
def financials(
    path: Path,
    countries: Annotated[List[CountriesEnum], typer.Argument()],
    filters: Optional[str] = None,
    api_keys: Optional[Path] = None,
) -> None:
    with console.status("[bold green]Fetching Financial data..."):
        source_api_keys = SourceApiKeys.from_file(api_keys)
        bearish = Bearish(path=path, api_keys=source_api_keys)
        filter = Filter(countries=countries, filters=filters)
        bearish.get_financials(filter)
        console.log("[bold][red]Financial data downloaded!")


@app.command()
def prices(
    path: Path,
    countries: Annotated[List[CountriesEnum], typer.Argument()],
    filters: Optional[str] = None,
    api_keys: Optional[Path] = None,
) -> None:
    with console.status("[bold green]Fetching Price data..."):
        source_api_keys = SourceApiKeys.from_file(api_keys)
        bearish = Bearish(path=path, api_keys=source_api_keys)
        filter = Filter(countries=countries, filters=filters)
        bearish.get_prices(filter)
        console.log("[bold][red]Price data downloaded!")


@app.command()
def update(
    path: Path,
    symbols: Optional[List[str]] = None,
    api_keys: Optional[Path] = None,
    series_length: str = "1mo",
) -> None:
    source_api_keys = SourceApiKeys.from_file(api_keys)
    bearish = Bearish(path=path, api_keys=source_api_keys)
    bearish.update_prices(symbols, series_length=series_length)  # type: ignore
    bearish.update_financials(symbols)


if __name__ == "__main__":
    app()



import logging
import time

from typing import List, Optional, Dict, Any, Callable

import pandas as pd
import yfinance as yf  # type: ignore
from pydantic import BaseModel, Field

from bearish.exchanges.exchanges import Countries
from bearish.models.base import Ticker
from bearish.models.query.query import AssetQuery
from bearish.models.assets.equity import Equity
from bearish.models.financials.balance_sheet import QuarterlyBalanceSheet
from bearish.models.financials.cash_flow import QuarterlyCashFlow
from bearish.models.financials.metrics import (
    QuarterlyFinancialMetrics,
)
from bearish.models.price.price import Price

from bearish.sources.base import (
    AbstractSource,
)
from bearish.models.financials.base import Financials
from bearish.models.assets.assets import Assets, FailedQueryAssets
from bearish.types import Sources, SeriesLength, DELAY
from yahooquery import Ticker as YahooQueryTicker  # type: ignore

from bearish.utils.utils import batch, safe_get

logger = logging.getLogger(__name__)


class YahooQueryBase(BaseModel):
    __source__: Sources = "YahooQuery"


class YahooQueryFinancialBase(YahooQueryBase):
    @classmethod
    def from_data_frame(
        cls,
        ticker_: str,
        data: pd.DataFrame,
    ) -> List["YahooQueryFinancialBase"]:
        return [
            cls.model_validate(data_ | {"symbol": ticker_})
            for data_ in data.to_dict(orient="records")
        ]


class YahooQueryAssetOutput(BaseModel):
    equities: List["YahooQueryAssetBase"]
    failed_query: List[Ticker] = Field(default_factory=list)


class YahooQueryAssetBase(YahooQueryBase):
    @classmethod
    def _from_tickers(
        cls, tickers: List[Ticker], function: Callable[[str, yf.Ticker], Dict[str, Any]]
    ) -> YahooQueryAssetOutput:
        equities = []
        failed_query: List[Ticker] = []
        chunks = batch(tickers, size=100)
        logger.debug(f"Retrieving {len(tickers)} assets.")
        for chunk in chunks:
            yahoo_tickers = YahooQueryTicker(
                " ".join([ticker.symbol for ticker in chunk])
            )
            asset_profile = yahoo_tickers.asset_profile
            summary_detail = yahoo_tickers.summary_detail
            summary_profile = yahoo_tickers.summary_profile
            key_stats = yahoo_tickers.key_stats
            financial_data = yahoo_tickers.financial_data
            quotes = yahoo_tickers.quotes
            for ticker in chunk:
                data = (
                    safe_get(asset_profile, ticker.symbol)
                    | safe_get(summary_detail, ticker.symbol)
                    | safe_get(summary_profile, ticker.symbol)
                    | safe_get(key_stats, ticker.symbol)
                    | safe_get(financial_data, ticker.symbol)
                    | safe_get(quotes, ticker.symbol)
                    | {"symbol": ticker.symbol}
                )
                equities.append(cls.model_validate(data))
            time.sleep(DELAY)  # Avoid hitting API rate limits
        logger.debug(f"Retrieved {len(equities)} assets.")
        return YahooQueryAssetOutput(equities=equities, failed_query=failed_query)


class YahooQueryEquity(YahooQueryAssetBase, Equity):
    __alias__ = {
        "city": "city",
        "zip": "zipcode",
        "country": "country",
        "website": "website",
        "industry": "industry",
        "industryKey": "industry_group",
        "sector": "sector",
        "dividendRate": "dividend_rate",
        "dividendYield": "dividend_yield",
        "marketCap": "market_capitalization",
        "currency": "currency",
        "market": "market",
        "profitMargins": "gross_margins",
        "floatShares": "float_shares",
        "sharesOutstanding": "shares_outstanding",
        "bookValue": "book_value",
        "priceToBook": "price_to_book",
        "earningsQuarterlyGrowth": "earning_growth",
        "trailingEps": "trailing_peg_ratio",
        "exchange": "exchange",
        "currentPrice": "current_ratio",
        "quickRatio": "quick_ratio",
        "revenuePerShare": "revenue_per_share",
        "returnOnAssets": "return_on_assets",
        "returnOnEquity": "return_on_equity",
        "operatingCashflow": "operating_margins",
        "revenueGrowth": "revenue_growth",
        "shortName": "short_shares",
        "longName": "name",
        "sourceInterval": "source",
        "symbol": "symbol",
    }

    @classmethod
    def from_tickers(cls, tickers: List[Ticker]) -> YahooQueryAssetOutput:
        return cls._from_tickers(tickers, lambda ticker, x: x.info)


class YahooQueryFinancialMetrics(YahooQueryFinancialBase, QuarterlyFinancialMetrics):
    __alias__ = {
        "symbol": "symbol",
        "asOfDate": "date",
        "EBITDA": "ebitda",
        "NetIncome": "net_income",
        "BasicEPS": "basic_eps",
        "DilutedEPS": "diluted_eps",
        "TotalRevenue": "total_revenue",
        "OperatingRevenue": "operating_revenue",
        "GrossProfit": "gross_profit",
        "TotalExpenses": "total_expenses",
        "OperatingIncome": "operating_income",
        "CostOfRevenue": "cost_of_revenue",
        "TaxProvision": "tax_provision",
        "TaxRateForCalcs": "tax_rate",
    }


class YahooQueryBalanceSheet(YahooQueryFinancialBase, QuarterlyBalanceSheet):
    __alias__ = {
        "symbol": "symbol",
        "asOfDate": "date",
        "TreasuryStock": "treasury_stock",
        "OrdinarySharesNumber": "common_stock_shares_outstanding",
        "CommonStock": "common_stock",
        "TotalDebt": "short_long_term_debt_total",
        "CapitalLeaseObligations": "capital_lease_obligations",
        "StockholdersEquity": "total_shareholder_equity",
        "RetainedEarnings": "retained_earnings",
        "TotalLiabilitiesNetMinorityInterest": "total_liabilities",
        "TotalNonCurrentLiabilitiesNetMinorityInterest": "total_non_current_liabilities",
        "OtherNonCurrentLiabilities": "other_non_current_liabilities",
        "LongTermDebt": "long_term_debt",
        "CurrentLiabilities": "total_current_liabilities",
        "OtherCurrentLiabilities": "other_current_liabilities",
        "CurrentDebt": "current_debt",
        "AccountsPayable": "current_accounts_payable",
        "TotalAssets": "total_assets",
        "TotalNonCurrentAssets": "total_non_current_assets",
        "OtherNonCurrentAssets": "other_non_current_assets",
        "AccumulatedDepreciation": "accumulated_depreciation_amortization_ppe",
        "GrossPPE": "property_plant_equipment",
        "CurrentAssets": "total_current_assets",
        "OtherCurrentAssets": "other_current_assets",
        "Inventory": "inventory",
        "AccountsReceivable": "current_net_receivables",
        "CashCashEquivalentsAndShortTermInvestments": "cash_and_short_term_investments",
        "CashAndCashEquivalents": "cash_and_cash_equivalents_at_carrying_value",
    }


class YahooQueryCashFlow(YahooQueryFinancialBase, QuarterlyCashFlow):
    __alias__ = {
        "symbol": "symbol",
        "asOfDate": "date",
        "OperatingCashFlow": "operating_cash_flow",
        "ChangeInOtherCurrentLiabilities": "change_in_operating_liabilities",
        "ChangeInWorkingCapital": "change_in_working_capital",
        "ChangeInOtherWorkingCapital": "change_in_other_working_capital",
        "ChangeInReceivables": "change_in_receivables",
        "ChangeInInventory": "change_in_inventory",
        "DepreciationAmortizationDepletion": "depreciation_amortization_depletion",
        "CapitalExpenditure": "capital_expenditure",
        "InvestingCashFlow": "cash_flow_from_investing_activities",
        "FinancingCashFlow": "financing_cash_flow",
        "RepurchaseOfCapitalStock": "repurchase_of_capital_stock",
        "CashDividendsPaid": "cash_dividends_paid",
        "CommonStockDividendPaid": "common_stock_dividend_paid",
        "CommonStockIssuance": "proceeds_from_issuance_of_common_stock",
        "ChangesInCash": "changes_in_cash",
        "NetIncomeFromContinuingOperations": "net_income_from_continuing_operations",
    }


class YahooQueryPrice(YahooQueryBase, Price):
    __alias__ = {
        "Open": "open",
        "High": "high",
        "Low": "low",
        "Close": "close",
        "Volume": "volume",
        "symbol": "symbol",
        "Date": "date",
    }


class YahooQuerySource(YahooQueryBase, AbstractSource):
    countries: List[Countries] = [
        "United Kingdom",
        "Germany",
        "France",
        "Netherlands",
        "Belgium",
        "US",
    ]

    def set_api_key(self, api_key: str) -> None: ...
    def _read_assets(self, query: Optional[AssetQuery] = None) -> Assets:
        if query is None:
            return Assets()

        if query.symbols.empty():
            return Assets()
        equities = YahooQueryEquity.from_tickers(query.symbols.equities)
        return Assets(
            equities=equities.equities,
            failed_query=FailedQueryAssets(symbols=equities.failed_query),
        )

    def _read_financials(self, tickers: List[str]) -> List[Financials]:
        financials = []
        yahoo_tickers = YahooQueryTicker(" ".join(tickers))
        cash_flow = yahoo_tickers.cash_flow()
        balance_sheet = yahoo_tickers.balance_sheet()
        income_statement = yahoo_tickers.income_statement()
        for ticker in tickers:
            current_cash_flow = cash_flow[cash_flow.index == ticker]
            current_balance_sheet = balance_sheet[balance_sheet.index == ticker]
            current_income_statement = income_statement[
                income_statement.index == ticker
            ]
            financials.append(
                Financials(
                    financial_metrics=YahooQueryFinancialMetrics.from_data_frame(
                        ticker, current_income_statement
                    ),
                    balance_sheets=YahooQueryBalanceSheet.from_data_frame(
                        ticker, current_balance_sheet
                    ),
                    cash_flows=YahooQueryCashFlow.from_data_frame(
                        ticker, current_cash_flow
                    ),
                )
            )
        return financials

    def _read_series(self, tickers: List[str], type: SeriesLength) -> List[Price]:
        yahoo_tickers = YahooQueryTicker(" ".join(tickers))
        data = yahoo_tickers.history(period=type)
        records = data.reset_index().to_dict(orient="records")
        return [YahooQueryPrice(**(record)) for record in records]



import json
from pathlib import Path
from typing import Dict, Optional

from pydantic import BaseModel, Field


class SourceApiKeys(BaseModel):
    keys: Dict[str, str] = Field(default_factory=dict)

    @classmethod
    def from_file(cls, api_keys_path: Optional[Path] = None) -> "SourceApiKeys":
        if api_keys_path is None:
            return cls()
        if not api_keys_path.exists():
            raise FileNotFoundError(f"File not found: {api_keys_path}")
        try:
            return cls(keys=json.loads(api_keys_path.read_text()))
        except Exception as e:
            raise ValueError(f"Invalid JSON in file {api_keys_path}: {e}") from e



import json
from typing import Optional, Annotated, Any, Dict

from pydantic import Field, BeforeValidator, model_validator

from bearish.models.base import DataSourceBase
from bearish.utils.utils import to_string, format_capitalize


class ComponentDescription(DataSourceBase):
    symbol: str = Field(
        description="Unique ticker symbol identifying the company on the stock exchange"
    )
    name: Annotated[
        Optional[str],
        BeforeValidator(to_string),
        Field(None, description="Full name of the company"),
    ]
    isin: Annotated[
        Optional[str],
        BeforeValidator(to_string),
        Field(
            default=None,
            description="International Securities Identification Number (ISIN) for the company's stock",
        ),
    ]


class BaseComponent(ComponentDescription):
    base_symbol: str = Field(
        description="Root symbol, primary ticker, or company identifier."
    )
    modifier: Optional[str] = Field(
        None, description="Suffix, modifier, attribute, or share class indicator."
    )
    summary: Annotated[
        Optional[str],
        BeforeValidator(to_string),
        Field(
            None, description="Brief summary of the company's operations and activities"
        ),
    ]
    currency: Annotated[
        Optional[str],
        BeforeValidator(to_string),
        Field(
            None,
            description="Currency code (e.g., USD, CNY) in which the company's financials are reported",
        ),
    ]
    exchange: Annotated[
        Optional[str],
        BeforeValidator(to_string),
        Field(
            None,
            description="Stock exchange where the company is listed, represented by its abbreviation",
        ),
    ]
    market: Annotated[
        Optional[str],
        BeforeValidator(to_string),
        Field(
            None,
            description="Market type or classification for the company's listing, such as 'Main Market'",
        ),
    ]
    country: Annotated[
        Optional[str],
        BeforeValidator(format_capitalize),
        Field(None, description="Country where the company's headquarters is located"),
    ]

    @model_validator(mode="before")
    @classmethod
    def _validator(cls, value: Dict[str, Any]) -> Dict[str, Any]:
        if "symbol" not in value:
            raise ValueError(
                f"Symbol is required for BaseComponent. Provided data {json.dumps(value, indent=4)}"
            )
        base_symbol, *modifier = value["symbol"].split(".")
        value["base_symbol"] = base_symbol
        value["modifier"] = modifier[0] if modifier else None
        return value



from math import isnan
from typing import Annotated, Optional

from pydantic import BeforeValidator, Field

from bearish.models.base import DataSourceBase
from bearish.utils.utils import to_float


class Price(DataSourceBase):
    open: Annotated[float, BeforeValidator(to_float), Field(None)]
    high: Annotated[float, BeforeValidator(to_float), Field(None)]
    low: Annotated[float, BeforeValidator(to_float), Field(None)]
    close: Annotated[float, BeforeValidator(to_float)]
    volume: Annotated[float, BeforeValidator(to_float), Field(None)]
    dividends: Annotated[Optional[float], BeforeValidator(to_float), Field(None)]
    stock_splits: Annotated[Optional[float], BeforeValidator(to_float), Field(None)]

    def valid(self) -> bool:
        return not any(
            isnan(field)
            for field in [self.open, self.high, self.low, self.close, self.volume]
        )



import datetime
import logging
from typing import List, Dict, Any, TYPE_CHECKING

import pandas as pd
from pydantic import BaseModel, Field

from bearish.models.base import Ticker, DataSourceBase
from bearish.models.financials.balance_sheet import BalanceSheet, QuarterlyBalanceSheet
from bearish.models.financials.cash_flow import CashFlow, QuarterlyCashFlow
from bearish.models.financials.earnings_date import EarningsDate
from bearish.models.financials.metrics import (
    FinancialMetrics,
    QuarterlyFinancialMetrics,
)
from bearish.models.query.query import AssetQuery, Symbols

if TYPE_CHECKING:
    from bearish.interface.interface import BearishDbBase
logger = logging.getLogger(__name__)


class Financials(BaseModel):
    financial_metrics: List[FinancialMetrics] = Field(default_factory=list)
    balance_sheets: List[BalanceSheet] = Field(default_factory=list)
    cash_flows: List[CashFlow] = Field(default_factory=list)
    quarterly_financial_metrics: List[QuarterlyFinancialMetrics] = Field(
        default_factory=list
    )
    quarterly_balance_sheets: List[QuarterlyBalanceSheet] = Field(default_factory=list)
    quarterly_cash_flows: List[QuarterlyCashFlow] = Field(default_factory=list)
    earnings_date: List[EarningsDate] = Field(default_factory=list)

    def _get_dates(self) -> List[datetime.date]:
        dates = sorted(
            {
                field_.date
                for field in self.model_fields
                for field_ in getattr(self, field)
            }
        )
        return dates

    def add(self, financials: "Financials") -> None:
        self.financial_metrics.extend(financials.financial_metrics)
        self.balance_sheets.extend(financials.balance_sheets)
        self.cash_flows.extend(financials.cash_flows)
        self.quarterly_financial_metrics.extend(financials.quarterly_financial_metrics)
        self.quarterly_balance_sheets.extend(financials.quarterly_balance_sheets)
        self.quarterly_cash_flows.extend(financials.quarterly_cash_flows)
        self.earnings_date.extend(financials.earnings_date)

    def is_empty(self) -> bool:
        return not any(
            [
                self.financial_metrics,
                self.balance_sheets,
                self.cash_flows,
                self.quarterly_financial_metrics,
                self.quarterly_balance_sheets,
                self.quarterly_cash_flows,
                self.earnings_date,
            ]
        )

    def _compute_growth(self, data: pd.DataFrame, field: str) -> Dict[str, Any]:
        data[f"{field}_growth"] = data[field].pct_change() * 100
        return {
            f"{field}_growth_{i}": value
            for i, value in enumerate(
                data[f"{field}_growth"].sort_index(ascending=False).tolist()
            )
        }

    @classmethod
    def from_ticker(cls, bearish_db: "BearishDbBase", ticker: Ticker) -> "Financials":
        return bearish_db.read_financials(
            AssetQuery(symbols=Symbols(equities=[ticker]))  # type: ignore
        )


class ManyFinancials(BaseModel):
    financials: List[Financials] = Field(default_factory=list)

    def get(self, attribute: str) -> List[DataSourceBase]:
        return [
            a
            for financial in self.financials
            if hasattr(financial, attribute)
            for a in getattr(financial, attribute)
        ]


class FinancialsWithDate(Financials):
    date: datetime.date

    @classmethod
    def from_financials(cls, financials_: Financials) -> List["FinancialsWithDate"]:
        financials_with_dates = []
        for date in financials_._get_dates():
            financials = cls(date=date)
            for field in financials_.model_fields:
                setattr(
                    financials,
                    field,
                    [
                        field_
                        for field_ in getattr(financials_, field)
                        if field_.date <= date
                    ],
                )
            financials_with_dates.append(financials)
        return financials_with_dates



from typing import Annotated, Optional

from pydantic import BeforeValidator, Field

from bearish.models.assets.base import BaseComponent
from bearish.utils.utils import to_string


class Currency(BaseComponent):
    base_currency: Annotated[
        Optional[str], BeforeValidator(to_string), Field(default=None)
    ]
    quote_currency: Annotated[
        Optional[str], BeforeValidator(to_string), Field(default=None)
    ]



import logging
import time
from datetime import date

from typing import List, Optional, Dict, Any, Callable, cast

import pandas as pd
import yfinance as yf  # type: ignore
from pydantic import BaseModel, Field
from tenacity import retry, stop_after_attempt, wait_fixed, wait_exponential

from bearish.exchanges.exchanges import Countries
from bearish.models.assets.etfs import Etf
from bearish.models.base import Ticker
from bearish.models.financials.earnings_date import EarningsDate
from bearish.models.query.query import AssetQuery
from bearish.models.assets.equity import Equity
from bearish.models.financials.balance_sheet import QuarterlyBalanceSheet
from bearish.models.financials.cash_flow import QuarterlyCashFlow
from bearish.models.financials.metrics import (
    QuarterlyFinancialMetrics,
)
from bearish.models.price.price import Price

from bearish.sources.base import (
    AbstractSource,
)
from bearish.models.financials.base import Financials
from bearish.models.assets.assets import Assets, FailedQueryAssets
from bearish.types import Sources, SeriesLength, DELAY

logger = logging.getLogger(__name__)


class YfinanceBase(BaseModel):
    __source__: Sources = "Yfinance"


@retry(
    wait=wait_exponential(multiplier=1, min=60, max=1200),
    stop=stop_after_attempt(5),
    reraise=True,
)
def get_data_frame(
    ticker_: yf.Ticker,
    attribute: str,
    transpose: bool = True,
    prefix: Optional[str] = None,
) -> pd.DataFrame:
    attribute_ = attribute if not prefix else f"{prefix}_{attribute}"
    data = getattr(ticker_, attribute_).T if transpose else getattr(ticker_, attribute_)
    if data is None:
        return pd.DataFrame()
    data.index = [date(index.year, index.month, index.day) for index in data.index]
    data = data.reset_index(names=["date"])
    return cast(pd.DataFrame, data)


class YfinanceFinancialBase(YfinanceBase):
    @classmethod
    def _from_ticker(
        cls,
        ticker_: yf.Ticker,
        attribute: str,
        transpose: bool = True,
        prefix: Optional[str] = None,
    ) -> List["YfinanceFinancialBase"]:
        try:
            data = get_data_frame(
                ticker_, attribute, transpose=transpose, prefix=prefix
            )
            time.sleep(DELAY)
            return [
                cls.model_validate(data_ | {"symbol": ticker_.ticker})
                for data_ in data.to_dict(orient="records")
            ]

        except Exception as e:
            logger.error(f"Error reading {ticker_.ticker} {attribute}: {e}")
            return []


@retry(stop=stop_after_attempt(2), wait=wait_fixed(10))
def get_info(
    ticker: Ticker, function: Callable[[str, yf.Ticker], Dict[str, Any]]
) -> Dict[str, Any]:
    current_ticker = yf.Ticker(ticker.symbol)
    info = function(ticker.symbol, current_ticker)
    return info


class YfinanceAssetOutput(BaseModel):
    equities: List["YfinanceAssetBase"]
    failed_query: List[Ticker] = Field(default_factory=list)


class YfinanceAssetBase(YfinanceBase):
    @classmethod
    def _from_tickers(
        cls, tickers: List[Ticker], function: Callable[[str, yf.Ticker], Dict[str, Any]]
    ) -> YfinanceAssetOutput:
        equities = []
        failed_query = []
        for ticker in tickers:
            try:
                info = get_info(ticker, function)
                if not info:
                    logger.error(f"No info found for {ticker.symbol}", exc_info=True)
                    failed_query.append(ticker)
                    continue
                logger.info(f"Successfully read {ticker.symbol}")
                equities.append(cls.model_validate(info))
            except Exception as e:
                logger.error(f"Error reading ticker: {e}", exc_info=True)
                failed_query.append(ticker)
                continue
            time.sleep(DELAY)  # To avoid hitting API limits
        return YfinanceAssetOutput(equities=equities, failed_query=failed_query)


class YfinanceEquity(YfinanceAssetBase, Equity):
    __alias__ = {
        "symbol": "symbol",
        "longName": "name",
        "longBusinessSummary": "summary",
        "currency": "currency",
        "exchange": "exchange",
        "sectorDisp": "sector",  # 'sectorDisp' seems like the descriptive sector field
        "industryDisp": "industry",  # 'industryDisp' matches the detailed industry description
        "sector": "sector",
        "industry": "industry",
        "industryKey": "industry_group",
        "country": "country",
        "state": "state",
        "city": "city",
        "zip": "zipcode",
        "website": "website",
        "marketCap": "market_capitalization",
        "sharesOutstanding": "shares_outstanding",
        "floatShares": "float_shares",
        "sharesShort": "short_shares",
        "bookValue": "book_value",
        "priceToBook": "price_to_book",
        "trailingPE": "trailing_price_to_earnings",
        "forwardPE": "forward_price_to_earnings",
        "dividendYield": "dividend_yield",
        "dividendRate": "dividend_rate",
        "trailingEps": "trailing_earnings_per_share",
        "forwardEps": "forward_earnings_per_share",
        "returnOnEquity": "return_on_equity",
        "operatingMargins": "operating_margins",
        "grossMargins": "gross_margins",
        "revenueGrowth": "revenue_growth",
        "revenuePerShare": "revenue_per_share",
        "quickRatio": "quick_ratio",
        "currentRatio": "current_ratio",
        "earningsGrowth": "earning_growth",
        "trailingPegRatio": "trailing_peg_ratio",
        "priceToSalesTrailing12Months": "trailing_price_to_sales",
        "returnOnAssets": "return_on_assets",
        "shortRatio": "short_ratio",
        "timeZone": "timezone",
        "isin": "isin",
        "cusip": "cusip",
        "figi": "figi",
        "compositeFigi": "composite_figi",
        "shareclassFigi": "shareclass_figi",
    }

    @classmethod
    def from_tickers(cls, tickers: List[Ticker]) -> YfinanceAssetOutput:
        return cls._from_tickers(tickers, lambda ticker, x: x.info)


def to_funds_data_dict(data: pd.DataFrame, ticker: str) -> Dict[str, Any]:
    data_dict = {}
    data_dict.update(data.to_dict().get(ticker, {}))
    return data_dict


def _get_etf(ticker: str, results: yf.Ticker) -> Dict[str, Any]:
    etf = {}
    etf.update(to_funds_data_dict(results.funds_data.fund_operations, ticker))
    etf.update(to_funds_data_dict(results.funds_data.equity_holdings, ticker))
    etf.update(results.funds_data.fund_overview)
    etf.update({"Sector Weightings": results.funds_data.sector_weightings})
    etf.update(
        {
            "Holding Percent": results.funds_data.top_holdings.to_dict().get(
                "Holding Percent", {}
            )
        }
    )
    etf.update({"summary": results.funds_data.description})
    etf.update({"isin": results.isin})
    etf.update(results.info)
    return etf


class YfinanceEtf(YfinanceAssetBase, Etf):
    __alias__ = {
        "symbol": "symbol",
        "Annual Report Expense Ratio": "annual_report_expense_ratio",
        "Annual Holdings Turnover": "annual_holdings_turnover",
        "Total Net Assets": "total_net_assets",
        "Price/Earnings": "price_to_earnings",
        "Price/Book": "price_to_book",
        "Price/Sales": "price_to_sales",
        "Price/Cashflow": "price_to_cashflow",
        "Median Market Cap": "median_market_cap",
        "3 Year Earnings Growth": "three_year_earnings_growth",
        "categoryName": "category",
        "family": "fund_family",
        "legalType": "legal_type",
        "Sector Weightings": "sector_weightings",
        "Holding Percent": "holding_percent",
        "summary": "long_business_summary",
        "isin": "isin",
        "longBusinessSummary": "long_business_summary",
        "trailingPE": "trailing_pe",
        "yield": "yield_",
        "navPrice": "nav_price",
        "currency": "currency",
        "category": "category",
        "ytdReturn": "ytd_return",
        "beta3Year": "beta_3_year",
        "fundFamily": "fund_family",
        "fundInceptionDate": "fund_inception_date",
        "threeYearAverageReturn": "three_year_average_return",
        "exchange": "exchange",
        "quoteType": "quote_type",
        "shortName": "short_name",
        "longName": "long_name",
    }

    @classmethod
    def from_tickers(cls, tickers: List[Ticker]) -> YfinanceAssetOutput:
        return cls._from_tickers(tickers, _get_etf)


class YfinanceFinancialMetrics(YfinanceFinancialBase, QuarterlyFinancialMetrics):
    __alias__ = {
        "symbol": "symbol",
        "EBITDA": "ebitda",
        "Net Income": "net_income",
        "Basic EPS": "basic_eps",
        "Diluted EPS": "diluted_eps",
        "Total Revenue": "total_revenue",
        "Operating Revenue": "operating_revenue",
        "Gross Profit": "gross_profit",
        "Total Expenses": "total_expenses",
        "Operating Income": "operating_income",
        "Cost Of Revenue": "cost_of_revenue",
        "Tax Provision": "tax_provision",
        "Tax Rate For Calcs": "tax_rate",
    }

    @classmethod
    def from_ticker(
        cls, ticker: yf.Ticker, prefix: Optional[str] = None
    ) -> List["YfinanceFinancialMetrics"]:
        return cls._from_ticker(ticker, "financials", prefix=prefix)  # type: ignore


class yFinanceEarningsDate(YfinanceFinancialBase, EarningsDate):
    __alias__ = {
        "symbol": "symbol",
        "EPS Estimate": "eps_estimate",
        "Reported EPS": "eps_reported",
    }

    @classmethod
    def from_ticker(
        cls, ticker: yf.Ticker, prefix: Optional[str] = None
    ) -> List["yFinanceEarningsDate"]:
        return cls._from_ticker(ticker, "earnings_dates", transpose=False, prefix=prefix)  # type: ignore


class yFinanceBalanceSheet(YfinanceFinancialBase, QuarterlyBalanceSheet):
    __alias__ = {
        "symbol": "symbol",
        "Treasury Shares Number": "treasury_stock",
        "Ordinary Shares Number": "common_stock_shares_outstanding",
        "Share Issued": "common_stock",
        "Total Debt": "short_long_term_debt_total",
        "Capital Lease Obligations": "capital_lease_obligations",
        "Common Stock Equity": "common_stock",  # Closest match
        "Total Equity Gross Minority Interest": "total_shareholder_equity",
        "Stockholders Equity": "total_shareholder_equity",
        "Retained Earnings": "retained_earnings",
        "Capital Stock": "common_stock",  # Closest match
        "Common Stock": "common_stock",
        "Total Liabilities Net Minority Interest": "total_liabilities",
        "Total Non Current Liabilities Net Minority Interest": "total_non_current_liabilities",
        "Other Non Current Liabilities": "other_non_current_liabilities",
        "Long Term Capital Lease Obligation": "capital_lease_obligations",  # Closest match
        "Long Term Debt": "long_term_debt",
        "Current Liabilities": "total_current_liabilities",
        "Other Current Liabilities": "other_current_liabilities",
        "Current Debt And Capital Lease Obligation": "current_debt",  # Closest match
        "Current Capital Lease Obligation": "capital_lease_obligations",  # Closest match
        "Current Debt": "current_debt",
        "Payables": "current_accounts_payable",  # Closest match
        "Accounts Payable": "current_accounts_payable",
        "Total Assets": "total_assets",
        "Total Non Current Assets": "total_non_current_assets",
        "Other Non Current Assets": "other_non_current_assets",
        "Net PPE": "property_plant_equipment",  # Closest match
        "Accumulated Depreciation": "accumulated_depreciation_amortization_ppe",
        "Gross PPE": "property_plant_equipment",
        "Current Assets": "total_current_assets",
        "Other Current Assets": "other_current_assets",
        "Inventory": "inventory",
        "Receivables": "current_net_receivables",
        "Accounts Receivable": "current_net_receivables",
        "Cash Cash Equivalents And Short Term Investments": "cash_and_short_term_investments",
        "Cash And Cash Equivalents": "cash_and_cash_equivalents_at_carrying_value",
        "Cash Equivalents": "cash_and_short_term_investments",  # Closest match
    }

    @classmethod
    def from_ticker(
        cls, ticker: yf.Ticker, prefix: Optional[str] = None
    ) -> List["yFinanceBalanceSheet"]:
        return cls._from_ticker(ticker, "balance_sheet", prefix=prefix)  # type: ignore


class yFinanceCashFlow(YfinanceFinancialBase, QuarterlyCashFlow):
    __alias__ = {
        "symbol": "symbol",
        "Operating Cash Flow": "operating_cash_flow",
        "Change In Payables And Accrued Expense": "change_in_operating_liabilities",  # Closest Match
        "Change In Working Capital": "change_in_working_capital",
        "Change In Other Working Capital": "change_in_other_working_capital",
        "Change In Receivables": "change_in_receivables",
        "Change In Inventory": "change_in_inventory",
        "Depreciation Amortization Depletion": "depreciation_amortization_depletion",
        "Capital Expenditure": "capital_expenditure",
        "Investing Cash Flow": "cash_flow_from_investing_activities",
        "Financing Cash Flow": "financing_cash_flow",
        "Repurchase Of Capital Stock": "repurchase_of_capital_stock",
        "Cash Dividends Paid": "cash_dividends_paid",
        "Common Stock Dividend Paid": "common_stock_dividend_paid",
        "Common Stock Issuance": "proceeds_from_issuance_of_common_stock",
        "Changes In Cash": "changes_in_cash",
        "Net Income From Continuing Operations": "net_income_from_continuing_operations",
    }

    @classmethod
    def from_ticker(
        cls, ticker: yf.Ticker, prefix: Optional[str] = None
    ) -> List["yFinanceCashFlow"]:
        return cls._from_ticker(ticker, "cashflow", prefix=prefix)  # type: ignore


class yFinancePrice(YfinanceBase, Price):
    __alias__ = {
        "Open": "open",
        "High": "high",
        "Low": "low",
        "Close": "close",
        "Volume": "volume",
        "symbol": "symbol",
        "Date": "date",
    }


class yFinanceSource(YfinanceBase, AbstractSource):
    countries: List[Countries] = [
        "United Kingdom",
        "Germany",
        "France",
        "Netherlands",
        "Belgium",
        "US",
    ]

    def set_api_key(self, api_key: str) -> None: ...
    def _read_assets(self, query: Optional[AssetQuery] = None) -> Assets:
        if query is None:
            return Assets()

        if query.symbols.empty():
            return Assets()
        equities = YfinanceEquity.from_tickers(query.symbols.equities)
        etfs = YfinanceEtf.from_tickers(query.symbols.etfs)
        return Assets(
            equities=equities.equities,
            etfs=etfs.equities,
            failed_query=FailedQueryAssets(
                symbols=etfs.failed_query + equities.failed_query
            ),
        )

    def _read_financials(self, tickers: List[str]) -> List[Financials]:
        financials = []
        for ticker in tickers:
            try:
                ticker_ = yf.Ticker(ticker)
            except Exception as e:
                logger.error(f"Error reading financials for {ticker}: {e}")
                continue
            financials.append(
                Financials(
                    financial_metrics=YfinanceFinancialMetrics.from_ticker(ticker_),
                    balance_sheets=yFinanceBalanceSheet.from_ticker(ticker_),
                    cash_flows=yFinanceCashFlow.from_ticker(ticker_),
                    quarterly_financial_metrics=YfinanceFinancialMetrics.from_ticker(
                        ticker_, prefix="quarterly"
                    ),
                    quarterly_balance_sheets=yFinanceBalanceSheet.from_ticker(
                        ticker_, prefix="quarterly"
                    ),
                    quarterly_cash_flows=yFinanceCashFlow.from_ticker(
                        ticker_, prefix="quarterly"
                    ),
                    earnings_date=yFinanceEarningsDate.from_ticker(ticker_),
                )
            )
        return financials

    def _read_series(  # type: ignore
        self, tickers: List[str], type: SeriesLength
    ) -> List[yFinancePrice]:
        time.sleep(5)
        data = yf.download(
            tickers, period=type, group_by="ticker", auto_adjust=True, timeout=60
        )
        records_final = []
        for ticker in tickers:
            if ticker in data.columns:
                records = data[ticker].reset_index().to_dict(orient="records")
                records_final.extend(
                    [
                        yFinancePrice(**(record | {"symbol": ticker}))
                        for record in records
                    ]
                )
        return records_final



from typing import Annotated, Optional

from pydantic import BeforeValidator, Field

from bearish.models.assets.base import BaseComponent
from bearish.utils.utils import to_string


class Crypto(BaseComponent):
    cryptocurrency: Annotated[
        Optional[str],
        BeforeValidator(to_string),
        Field(None, description="Name of the cryptocurrency"),
    ]



import logging
from datetime import date
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
from pydantic import BaseModel

from bearish.models.base import Ticker
from bearish.models.price.price import Price
from bearish.models.query.query import AssetQuery, Symbols
from bearish.utils.utils import to_dataframe


if TYPE_CHECKING:
    from bearish.interface.interface import BearishDbBase
logger = logging.getLogger(__name__)


class Prices(BaseModel):
    prices: list[Price]

    def get_last_date(self) -> date:
        return sorted(self.prices, key=lambda price: price.date)[-1].date

    @classmethod
    def from_ticker(cls, bearish_db: "BearishDbBase", ticker: Ticker) -> "Prices":
        prices = bearish_db.read_series(
            AssetQuery(symbols=Symbols(equities=[ticker])), months=12 * 8  # type: ignore
        )
        return cls(prices=prices)

    def to_dataframe(self) -> pd.DataFrame:
        return to_dataframe(self.prices)

    def to_csv(self, json_path: Path | str) -> None:
        self.to_dataframe().to_csv(json_path)

    @classmethod
    def from_csv(cls, json_path: Path | str) -> "Prices":
        data = pd.read_csv(json_path)
        data["date"] = pd.to_datetime(data["date"], utc=True)
        return cls(
            prices=[
                Price.model_validate(d)
                for d in data.replace(np.nan, None).to_dict(orient="records")
            ]
        )



from datetime import datetime
from typing import Optional, Dict

from sqlalchemy import JSON, Column
from sqlmodel import SQLModel, Field


from bearish.models.assets.equity import Equity
from bearish.models.assets.crypto import Crypto
from bearish.models.assets.currency import Currency
from bearish.models.assets.etfs import Etf
from bearish.models.base import PriceTracker, FinancialsTracker
from bearish.models.financials.balance_sheet import BalanceSheet, QuarterlyBalanceSheet
from bearish.models.financials.cash_flow import CashFlow, QuarterlyCashFlow
from bearish.models.financials.earnings_date import EarningsDate
from bearish.models.financials.metrics import (
    FinancialMetrics,
    QuarterlyFinancialMetrics,
)
from bearish.models.price.price import Price


class BaseBearishTable(SQLModel):
    symbol: str = Field(index=True)
    source: str = Field(index=True)


class BaseTable(BaseBearishTable):
    __table_args__ = {"sqlite_autoincrement": True}
    id: Optional[int] = Field(default=None, primary_key=True)


class BaseFinancials(SQLModel):
    date: datetime = Field(primary_key=True, index=True)
    symbol: str = Field(primary_key=True, index=True)
    source: str = Field(primary_key=True, index=True)


class EquityORM(BaseTable, Equity, table=True):  # type: ignore
    __tablename__ = "equity"
    country: Optional[str] = Field(default=None, index=True)


class CryptoORM(BaseTable, Crypto, table=True):  # type: ignore
    __tablename__ = "crypto"
    cryptocurrency: Optional[str] = Field(default=None, index=True)


class CurrencyORM(BaseTable, Currency, table=True):  # type: ignore
    __tablename__ = "currency"
    base_currency: str = Field(index=True)


class EtfORM(BaseTable, Etf, table=True):  # type: ignore
    __tablename__ = "etf"

    holding_percent: Optional[Dict[str, float]] = Field(None, sa_column=Column(JSON))
    sector_weightings: Optional[Dict[str, float]] = Field(None, sa_column=Column(JSON))


class PriceORM(SQLModel, Price, table=True):  # type: ignore
    __tablename__ = "price"
    date: datetime = Field(primary_key=True, index=True)
    symbol: str = Field(primary_key=True, index=True)
    source: str = Field(primary_key=True, index=True)  # type: ignore


class FinancialMetricsORM(BaseFinancials, FinancialMetrics, table=True):  # type: ignore
    __tablename__ = "financialmetrics"


class BalanceSheetORM(BaseFinancials, BalanceSheet, table=True):  # type: ignore
    __tablename__ = "balancesheet"


class CashFlowORM(BaseFinancials, CashFlow, table=True):  # type: ignore
    __tablename__ = "cashflow"


class QuarterlyFinancialMetricsORM(BaseFinancials, QuarterlyFinancialMetrics, table=True):  # type: ignore
    __tablename__ = "quarterlyfinancialmetrics"


class QuarterlyBalanceSheetORM(BaseFinancials, QuarterlyBalanceSheet, table=True):  # type: ignore
    __tablename__ = "quarterlybalancesheet"


class QuarterlyCashFlowORM(BaseFinancials, QuarterlyCashFlow, table=True):  # type: ignore
    __tablename__ = "quarterlycashflow"


class EarningsDateORM(BaseFinancials, EarningsDate, table=True):  # type: ignore
    __tablename__ = "earningsdate"


class SourcesORM(SQLModel, table=True):
    __tablename__ = "sources"
    source: str = Field(primary_key=True, index=True)


class PriceTrackerORM(SQLModel, PriceTracker, table=True):
    __tablename__ = "pricetracker"
    __table_args__ = {"sqlite_autoincrement": True}
    source: str = Field(index=True, primary_key=True)
    symbol: str = Field(index=True, primary_key=True)


class FinancialsTrackerORM(SQLModel, FinancialsTracker, table=True):
    __tablename__ = "financialstracker"
    __table_args__ = {"sqlite_autoincrement": True}
    source: str = Field(index=True, primary_key=True)
    symbol: str = Field(index=True, primary_key=True)



from typing import List, Literal, Optional

from pydantic import BaseModel, Field, model_validator, validate_call

from bearish.models.base import Ticker
from bearish.models.query.query import AssetQuery
from bearish.types import Sources

Countries = Literal[
    "US",
    "Canada",
    "United Kingdom",
    "Germany",
    "France",
    "Netherlands",
    "Belgium",
    "Italy",
    "Spain",
    "Switzerland",
    "Sweden",
    "Denmark",
    "Norway",
    "Finland",
    "Portugal",
    "Austria",
    "Australia",
    "New Zealand",
    "Japan",
    "China",
    "Hong Kong",
    "Singapore",
    "India",
    "South Korea",
    "Taiwan",
    "Brazil",
    "Mexico",
    "Argentina",
    "Russia",
    "South Africa",
]

ExchangeType = Literal["suffixes", "aliases"]


class ExchangeQuery(BaseModel):
    suffixes: List[str] = Field(default_factory=list)
    aliases: List[str] = Field(default_factory=list)
    sources: List[str] = Field(default_factory=list)

    def to_suffixes_sql_statement(self) -> str:
        return f"""({' OR '.join([f"symbol LIKE '%{suffix}'" for suffix in self.suffixes])})"""

    def to_aliases_sql_statement(self) -> str:
        return f"""({','.join([f"'{alias}'" for alias in self.aliases])})"""

    def to_sources_sql_statement(self) -> str:
        if not self.sources:
            return ""
        return f"""({','.join([f"'{source}'" for source in self.sources])})"""

    def included(self, ticker: Ticker) -> bool:
        return (
            any(ticker.symbol.endswith(suffix) for suffix in self.suffixes)
            or ticker.exchange in self.aliases
        )


class Exchange(ExchangeQuery):
    name: str

    @model_validator(mode="after")
    def _model_validator(self) -> "Exchange":
        if not self.aliases and not self.suffixes:
            raise ValueError("At least one of names or suffixes must be provided")
        return self


class CountryExchanges(BaseModel):
    country: Countries
    exchanges: List[Exchange]


class Exchanges(BaseModel):
    exchanges: List[CountryExchanges] = Field(default_factory=list)

    @validate_call
    def get_exchanges(
        self, countries: List[Countries], type: ExchangeType = "suffixes"
    ) -> List[str]:
        return [
            suffix
            for country in countries
            for suffix in self._get_exchanges(country, type)
        ]

    def _get_exchanges(
        self, country: Countries, type: ExchangeType = "suffixes"
    ) -> List[str]:
        return [
            suffix
            for country_exchange in self.exchanges
            for exchange in country_exchange.exchanges
            for suffix in getattr(exchange, type)
            if country_exchange.country == country
        ]

    @validate_call
    def get_exchange_query(
        self, countries: List[Countries], sources: Optional[List[Sources]] = None
    ) -> ExchangeQuery:
        sources = sources or []
        suffixes = self.get_exchanges(countries, "suffixes")
        aliases = self.get_exchanges(countries, "aliases")
        return ExchangeQuery(suffixes=suffixes, aliases=aliases, sources=sources)

    def ticker_belongs_to_countries(
        self, ticker: Ticker, countries: List[Countries]
    ) -> bool:
        exchange_query = self.get_exchange_query(countries)
        return exchange_query.included(ticker)

    def get_asset_query(
        self, asset_query: AssetQuery, countries: List[Countries]
    ) -> AssetQuery:
        exchange_query = self.get_exchange_query(countries)
        symbols = asset_query.symbols.filter(exchange_query.included)
        asset_query_ = AssetQuery.model_validate(asset_query.model_dump())
        asset_query_.symbols = symbols
        return asset_query_


def exchanges_factory() -> Exchanges:
    return Exchanges(
        exchanges=[
            CountryExchanges(
                country="US",
                exchanges=[
                    Exchange(
                        name="NASDAQ",
                        suffixes=[".OQ"],
                        aliases=[
                            "NMS",
                            "NCM",
                            "NAS",
                            "NASDAQ",
                            "Nasdaq",
                            "NASDAQ Global Market",
                            "NASDAQ Capital Market",
                            "NASDAQ Global Select",
                            "NASDAQ Capital Markets",
                            "NASDAQ Stock Exchange",
                            "NASDAQ Stock Market",
                        ],
                    ),
                    Exchange(
                        name="New York Stock Exchange",
                        suffixes=[".N"],
                        aliases=[
                            "NYQ",
                            "NYS",
                            "New York Stock Exchange",
                            "New York Stock Exchange Arca",
                            "NYSE",
                            "NYSE American",
                            "American Stock Exchange",
                            "AMEX",
                            "NYSE Arca",
                        ],
                    ),
                    Exchange(name="NYSE American", suffixes=[".A"]),
                ],
            ),
            CountryExchanges(
                country="Canada",
                exchanges=[
                    Exchange(name="Toronto Stock Exchange", suffixes=[".TO"]),
                    Exchange(name="TSX Venture Exchange", suffixes=[".V"]),
                ],
            ),
            CountryExchanges(
                country="United Kingdom",
                exchanges=[Exchange(name="London Stock Exchange", suffixes=[".L"])],
            ),
            CountryExchanges(
                country="Germany",
                exchanges=[
                    Exchange(name="Deutsche Börse Xetra", suffixes=[".DE"]),
                    Exchange(name="Frankfurt Stock Exchange", suffixes=[".F"]),
                    Exchange(name="Munich Stock Exchange", suffixes=[".MU"]),
                    Exchange(name="Stuttgart Stock Exchange", suffixes=[".SG"]),
                ],
            ),
            CountryExchanges(
                country="France",
                exchanges=[Exchange(name="Euronext Paris", suffixes=[".PA"])],
            ),
            CountryExchanges(
                country="Netherlands",
                exchanges=[Exchange(name="Euronext Amsterdam", suffixes=[".AS"])],
            ),
            CountryExchanges(
                country="Belgium",
                exchanges=[Exchange(name="Euronext Brussels", suffixes=[".BR"])],
            ),
            CountryExchanges(
                country="Italy",
                exchanges=[Exchange(name="Borsa Italiana", suffixes=[".MI"])],
            ),
            CountryExchanges(
                country="Spain",
                exchanges=[Exchange(name="Bolsa de Madrid", suffixes=[".MC"])],
            ),
            CountryExchanges(
                country="Switzerland",
                exchanges=[Exchange(name="SIX Swiss Exchange", suffixes=[".SW"])],
            ),
            CountryExchanges(
                country="Sweden",
                exchanges=[Exchange(name="Stockholm Stock Exchange", suffixes=[".ST"])],
            ),
            CountryExchanges(
                country="Denmark",
                exchanges=[
                    Exchange(name="Copenhagen Stock Exchange", suffixes=[".CO"])
                ],
            ),
            CountryExchanges(
                country="Norway",
                exchanges=[Exchange(name="Oslo Stock Exchange", suffixes=[".OL"])],
            ),
            CountryExchanges(
                country="Finland",
                exchanges=[Exchange(name="Helsinki Stock Exchange", suffixes=[".HE"])],
            ),
            CountryExchanges(
                country="Portugal",
                exchanges=[Exchange(name="Euronext Lisbon", suffixes=[".LS"])],
            ),
            CountryExchanges(
                country="Austria",
                exchanges=[Exchange(name="Vienna Stock Exchange", suffixes=[".VI"])],
            ),
            CountryExchanges(
                country="Australia",
                exchanges=[
                    Exchange(name="Australian Securities Exchange", suffixes=[".AX"])
                ],
            ),
            CountryExchanges(
                country="New Zealand",
                exchanges=[Exchange(name="New Zealand Exchange", suffixes=[".NZ"])],
            ),
            CountryExchanges(
                country="Japan",
                exchanges=[Exchange(name="Tokyo Stock Exchange", suffixes=[".T"])],
            ),
            CountryExchanges(
                country="China",
                exchanges=[
                    Exchange(name="Shanghai Stock Exchange", suffixes=[".SS"]),
                    Exchange(name="Shenzhen Stock Exchange", suffixes=[".SZ"]),
                ],
            ),
            CountryExchanges(
                country="Hong Kong",
                exchanges=[Exchange(name="Hong Kong Stock Exchange", suffixes=[".HK"])],
            ),
            CountryExchanges(
                country="Singapore",
                exchanges=[Exchange(name="Singapore Exchange", suffixes=[".SI"])],
            ),
            CountryExchanges(
                country="India",
                exchanges=[
                    Exchange(name="National Stock Exchange", suffixes=[".NS"]),
                    Exchange(name="Bombay Stock Exchange", suffixes=[".BO"]),
                ],
            ),
            CountryExchanges(
                country="South Korea",
                exchanges=[Exchange(name="Korea Exchange", suffixes=[".KS"])],
            ),
            CountryExchanges(
                country="Taiwan",
                exchanges=[Exchange(name="Taiwan Stock Exchange", suffixes=[".TW"])],
            ),
            CountryExchanges(
                country="Brazil",
                exchanges=[
                    Exchange(name="B3 - Brazil Stock Exchange", suffixes=[".SA"])
                ],
            ),
            CountryExchanges(
                country="Mexico",
                exchanges=[Exchange(name="Mexican Stock Exchange", suffixes=[".MX"])],
            ),
            CountryExchanges(
                country="Argentina",
                exchanges=[
                    Exchange(name="Buenos Aires Stock Exchange", suffixes=[".BA"])
                ],
            ),
            CountryExchanges(
                country="Russia",
                exchanges=[Exchange(name="Moscow Exchange", suffixes=[".ME"])],
            ),
            CountryExchanges(
                country="South Africa",
                exchanges=[
                    Exchange(name="Johannesburg Stock Exchange", suffixes=[".JO"])
                ],
            ),
        ]
    )



"""price growth

Revision ID: 5f6984590216
Revises: 
Create Date: 2025-06-11 21:41:22.280529

"""

from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
import sqlmodel

# revision identifiers, used by Alembic.
revision: str = "5f6984590216"
down_revision: Union[str, None] = None
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "analysis",
        sa.Column("quarterly_positive_free_cash_flow", sa.Float(), nullable=True),
        sa.Column("quarterly_growing_operating_cash_flow", sa.Float(), nullable=True),
        sa.Column(
            "quarterly_operating_cash_flow_is_higher_than_net_income",
            sa.Float(),
            nullable=True,
        ),
        sa.Column("quarterly_mean_capex_ratio", sa.Float(), nullable=True),
        sa.Column("quarterly_max_capex_ratio", sa.Float(), nullable=True),
        sa.Column("quarterly_min_capex_ratio", sa.Float(), nullable=True),
        sa.Column("quarterly_mean_dividend_payout_ratio", sa.Float(), nullable=True),
        sa.Column("quarterly_max_dividend_payout_ratio", sa.Float(), nullable=True),
        sa.Column("quarterly_min_dividend_payout_ratio", sa.Float(), nullable=True),
        sa.Column("quarterly_positive_net_income", sa.Float(), nullable=True),
        sa.Column("quarterly_positive_operating_income", sa.Float(), nullable=True),
        sa.Column("quarterly_growing_net_income", sa.Float(), nullable=True),
        sa.Column("quarterly_growing_operating_income", sa.Float(), nullable=True),
        sa.Column("quarterly_positive_diluted_eps", sa.Float(), nullable=True),
        sa.Column("quarterly_positive_basic_eps", sa.Float(), nullable=True),
        sa.Column("quarterly_growing_basic_eps", sa.Float(), nullable=True),
        sa.Column("quarterly_growing_diluted_eps", sa.Float(), nullable=True),
        sa.Column("quarterly_positive_debt_to_equity", sa.Float(), nullable=True),
        sa.Column("quarterly_positive_return_on_assets", sa.Float(), nullable=True),
        sa.Column("quarterly_positive_return_on_equity", sa.Float(), nullable=True),
        sa.Column("quarterly_earning_per_share", sa.Float(), nullable=True),
        sa.Column("positive_free_cash_flow", sa.Float(), nullable=True),
        sa.Column("growing_operating_cash_flow", sa.Float(), nullable=True),
        sa.Column(
            "operating_cash_flow_is_higher_than_net_income", sa.Float(), nullable=True
        ),
        sa.Column("mean_capex_ratio", sa.Float(), nullable=True),
        sa.Column("max_capex_ratio", sa.Float(), nullable=True),
        sa.Column("min_capex_ratio", sa.Float(), nullable=True),
        sa.Column("mean_dividend_payout_ratio", sa.Float(), nullable=True),
        sa.Column("max_dividend_payout_ratio", sa.Float(), nullable=True),
        sa.Column("min_dividend_payout_ratio", sa.Float(), nullable=True),
        sa.Column("positive_net_income", sa.Float(), nullable=True),
        sa.Column("positive_operating_income", sa.Float(), nullable=True),
        sa.Column("growing_net_income", sa.Float(), nullable=True),
        sa.Column("growing_operating_income", sa.Float(), nullable=True),
        sa.Column("positive_diluted_eps", sa.Float(), nullable=True),
        sa.Column("positive_basic_eps", sa.Float(), nullable=True),
        sa.Column("growing_basic_eps", sa.Float(), nullable=True),
        sa.Column("growing_diluted_eps", sa.Float(), nullable=True),
        sa.Column("positive_debt_to_equity", sa.Float(), nullable=True),
        sa.Column("positive_return_on_assets", sa.Float(), nullable=True),
        sa.Column("positive_return_on_equity", sa.Float(), nullable=True),
        sa.Column("earning_per_share", sa.Float(), nullable=True),
        sa.Column("rsi_last_value", sa.Float(), nullable=True),
        sa.Column("macd_12_26_9_buy_date", sa.Date(), nullable=True),
        sa.Column("ma_50_200_buy_date", sa.Date(), nullable=True),
        sa.Column("slope_7", sa.Float(), nullable=True),
        sa.Column("slope_14", sa.Float(), nullable=True),
        sa.Column("slope_30", sa.Float(), nullable=True),
        sa.Column("slope_60", sa.Float(), nullable=True),
        sa.Column("last_adx", sa.Float(), nullable=True),
        sa.Column("last_dmp", sa.Float(), nullable=True),
        sa.Column("last_dmn", sa.Float(), nullable=True),
        sa.Column("last_price", sa.Float(), nullable=True),
        sa.Column("last_price_date", sa.Date(), nullable=True),
        sa.Column("year_to_date_growth", sa.Float(), nullable=True),
        sa.Column("last_52_weeks_growth", sa.Float(), nullable=True),
        sa.Column("last_week_growth", sa.Float(), nullable=True),
        sa.Column("last_month_growth", sa.Float(), nullable=True),
        sa.Column("last_year_growth", sa.Float(), nullable=True),
        sa.Column("year_to_date_max_growth", sa.Float(), nullable=True),
        sa.Column("last_week_max_growth", sa.Float(), nullable=True),
        sa.Column("last_month_max_growth", sa.Float(), nullable=True),
        sa.Column("last_year_max_growth", sa.Float(), nullable=True),
        sa.Column("macd_12_26_9_buy", sa.Float(), nullable=True),
        sa.Column("star_yoy", sa.Float(), nullable=True),
        sa.Column("star_wow", sa.Float(), nullable=True),
        sa.Column("star_mom", sa.Float(), nullable=True),
        sa.Column("symbol", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("exchange", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("source", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("date", sa.Date(), nullable=False),
        sa.Column("created_at", sa.Date(), nullable=False),
        sa.Column("name", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("isin", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("base_symbol", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("modifier", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("summary", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("currency", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("market", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("country", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("sector", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("industry_group", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("industry", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("website", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("market_capitalization", sa.Float(), nullable=True),
        sa.Column("book_value", sa.Float(), nullable=True),
        sa.Column("price_to_book", sa.Float(), nullable=True),
        sa.Column("trailing_price_to_earnings", sa.Float(), nullable=True),
        sa.Column("dividend_yield", sa.Float(), nullable=True),
        sa.Column("dividend_rate", sa.Float(), nullable=True),
        sa.Column("trailing_earnings_per_share", sa.Float(), nullable=True),
        sa.Column("forward_earnings_per_share", sa.Float(), nullable=True),
        sa.Column("return_on_equity", sa.Float(), nullable=True),
        sa.Column("operating_margins", sa.Float(), nullable=True),
        sa.Column("gross_margins", sa.Float(), nullable=True),
        sa.Column("revenue_growth", sa.Float(), nullable=True),
        sa.Column("price_per_earning_ratio", sa.Float(), nullable=True),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sqlite_autoincrement=True,
    )
    with op.batch_alter_table("analysis", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_analysis_source"), ["source"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_analysis_symbol"), ["symbol"], unique=False
        )

    op.create_table(
        "balancesheet",
        sa.Column("symbol", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("exchange", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("source", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("date", sa.DateTime(), nullable=False),
        sa.Column("created_at", sa.Date(), nullable=False),
        sa.Column("treasury_stock", sa.Float(), nullable=True),
        sa.Column("common_stock_shares_outstanding", sa.Float(), nullable=True),
        sa.Column("common_stock", sa.Float(), nullable=True),
        sa.Column("short_long_term_debt_total", sa.Float(), nullable=True),
        sa.Column("capital_lease_obligations", sa.Float(), nullable=True),
        sa.Column("total_shareholder_equity", sa.Float(), nullable=True),
        sa.Column("retained_earnings", sa.Float(), nullable=True),
        sa.Column("total_liabilities", sa.Float(), nullable=True),
        sa.Column("total_non_current_liabilities", sa.Float(), nullable=True),
        sa.Column("other_non_current_liabilities", sa.Float(), nullable=True),
        sa.Column("long_term_debt", sa.Float(), nullable=True),
        sa.Column("total_current_liabilities", sa.Float(), nullable=True),
        sa.Column("other_current_liabilities", sa.Float(), nullable=True),
        sa.Column("current_debt", sa.Float(), nullable=True),
        sa.Column("current_accounts_payable", sa.Float(), nullable=True),
        sa.Column("total_assets", sa.Float(), nullable=True),
        sa.Column("total_non_current_assets", sa.Float(), nullable=True),
        sa.Column("other_non_current_assets", sa.Float(), nullable=True),
        sa.Column(
            "accumulated_depreciation_amortization_ppe", sa.Float(), nullable=True
        ),
        sa.Column("property_plant_equipment", sa.Float(), nullable=True),
        sa.Column("total_current_assets", sa.Float(), nullable=True),
        sa.Column("other_current_assets", sa.Float(), nullable=True),
        sa.Column("inventory", sa.Float(), nullable=True),
        sa.Column("current_net_receivables", sa.Float(), nullable=True),
        sa.Column("cash_and_short_term_investments", sa.Float(), nullable=True),
        sa.Column(
            "cash_and_cash_equivalents_at_carrying_value", sa.Float(), nullable=True
        ),
        sa.PrimaryKeyConstraint("symbol", "source", "date"),
    )
    with op.batch_alter_table("balancesheet", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_balancesheet_date"), ["date"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_balancesheet_source"), ["source"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_balancesheet_symbol"), ["symbol"], unique=False
        )

    op.create_table(
        "cashflow",
        sa.Column("symbol", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("exchange", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("source", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("date", sa.DateTime(), nullable=False),
        sa.Column("created_at", sa.Date(), nullable=False),
        sa.Column("operating_cash_flow", sa.Float(), nullable=True),
        sa.Column("change_in_operating_liabilities", sa.Float(), nullable=True),
        sa.Column("change_in_working_capital", sa.Float(), nullable=True),
        sa.Column("change_in_other_working_capital", sa.Float(), nullable=True),
        sa.Column("change_in_receivables", sa.Float(), nullable=True),
        sa.Column("change_in_inventory", sa.Float(), nullable=True),
        sa.Column("depreciation_amortization_depletion", sa.Float(), nullable=True),
        sa.Column("capital_expenditure", sa.Float(), nullable=True),
        sa.Column("cash_flow_from_investing_activities", sa.Float(), nullable=True),
        sa.Column("financing_cash_flow", sa.Float(), nullable=True),
        sa.Column("repurchase_of_capital_stock", sa.Float(), nullable=True),
        sa.Column("cash_dividends_paid", sa.Float(), nullable=True),
        sa.Column("common_stock_dividend_paid", sa.Float(), nullable=True),
        sa.Column("proceeds_from_issuance_of_common_stock", sa.Float(), nullable=True),
        sa.Column("changes_in_cash", sa.Float(), nullable=True),
        sa.Column("net_income_from_continuing_operations", sa.Float(), nullable=True),
        sa.PrimaryKeyConstraint("symbol", "source", "date"),
    )
    with op.batch_alter_table("cashflow", schema=None) as batch_op:
        batch_op.create_index(batch_op.f("ix_cashflow_date"), ["date"], unique=False)
        batch_op.create_index(
            batch_op.f("ix_cashflow_source"), ["source"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_cashflow_symbol"), ["symbol"], unique=False
        )

    op.create_table(
        "crypto",
        sa.Column("symbol", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("exchange", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("source", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("date", sa.Date(), nullable=False),
        sa.Column("created_at", sa.Date(), nullable=False),
        sa.Column("name", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("isin", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("base_symbol", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("modifier", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("summary", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("currency", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("market", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("country", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("cryptocurrency", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sqlite_autoincrement=True,
    )
    with op.batch_alter_table("crypto", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_crypto_cryptocurrency"), ["cryptocurrency"], unique=False
        )
        batch_op.create_index(batch_op.f("ix_crypto_source"), ["source"], unique=False)
        batch_op.create_index(batch_op.f("ix_crypto_symbol"), ["symbol"], unique=False)

    op.create_table(
        "currency",
        sa.Column("symbol", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("exchange", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("source", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("date", sa.Date(), nullable=False),
        sa.Column("created_at", sa.Date(), nullable=False),
        sa.Column("name", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("isin", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("base_symbol", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("modifier", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("summary", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("currency", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("market", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("country", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("base_currency", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("quote_currency", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sqlite_autoincrement=True,
    )
    with op.batch_alter_table("currency", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_currency_base_currency"), ["base_currency"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_currency_source"), ["source"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_currency_symbol"), ["symbol"], unique=False
        )

    op.create_table(
        "earningsdate",
        sa.Column("symbol", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("exchange", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("source", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("date", sa.DateTime(), nullable=False),
        sa.Column("created_at", sa.Date(), nullable=False),
        sa.Column("eps_estimate", sa.Float(), nullable=True),
        sa.Column("eps_reported", sa.Float(), nullable=True),
        sa.Column("revenue_estimate", sa.Float(), nullable=True),
        sa.Column("revenue_reported", sa.Float(), nullable=True),
        sa.PrimaryKeyConstraint("symbol", "source", "date"),
    )
    with op.batch_alter_table("earningsdate", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_earningsdate_date"), ["date"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_earningsdate_source"), ["source"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_earningsdate_symbol"), ["symbol"], unique=False
        )

    op.create_table(
        "equity",
        sa.Column("symbol", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("exchange", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("source", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("date", sa.Date(), nullable=False),
        sa.Column("created_at", sa.Date(), nullable=False),
        sa.Column("name", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("isin", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("base_symbol", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("modifier", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("summary", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("currency", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("market", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("country", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("sector", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("industry_group", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("industry", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("website", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("market_capitalization", sa.Float(), nullable=True),
        sa.Column("book_value", sa.Float(), nullable=True),
        sa.Column("price_to_book", sa.Float(), nullable=True),
        sa.Column("trailing_price_to_earnings", sa.Float(), nullable=True),
        sa.Column("dividend_yield", sa.Float(), nullable=True),
        sa.Column("dividend_rate", sa.Float(), nullable=True),
        sa.Column("trailing_earnings_per_share", sa.Float(), nullable=True),
        sa.Column("forward_earnings_per_share", sa.Float(), nullable=True),
        sa.Column("return_on_equity", sa.Float(), nullable=True),
        sa.Column("operating_margins", sa.Float(), nullable=True),
        sa.Column("gross_margins", sa.Float(), nullable=True),
        sa.Column("revenue_growth", sa.Float(), nullable=True),
        sa.Column("state", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("city", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("zipcode", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("market_cap", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("shares_outstanding", sa.Float(), nullable=True),
        sa.Column("float_shares", sa.Float(), nullable=True),
        sa.Column("short_shares", sa.Float(), nullable=True),
        sa.Column("forward_price_to_earnings", sa.Float(), nullable=True),
        sa.Column("revenue_per_share", sa.Float(), nullable=True),
        sa.Column("quick_ratio", sa.Float(), nullable=True),
        sa.Column("current_ratio", sa.Float(), nullable=True),
        sa.Column("earning_growth", sa.Float(), nullable=True),
        sa.Column("trailing_peg_ratio", sa.Float(), nullable=True),
        sa.Column("trailing_price_to_sales", sa.Float(), nullable=True),
        sa.Column("return_on_assets", sa.Float(), nullable=True),
        sa.Column("short_ratio", sa.Float(), nullable=True),
        sa.Column("timezone", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("cusip", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("figi", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("composite_figi", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("shareclass_figi", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sqlite_autoincrement=True,
    )
    with op.batch_alter_table("equity", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_equity_country"), ["country"], unique=False
        )
        batch_op.create_index(batch_op.f("ix_equity_source"), ["source"], unique=False)
        batch_op.create_index(batch_op.f("ix_equity_symbol"), ["symbol"], unique=False)

    op.create_table(
        "etf",
        sa.Column("symbol", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("exchange", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("source", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("date", sa.Date(), nullable=False),
        sa.Column("created_at", sa.Date(), nullable=False),
        sa.Column("name", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("isin", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("base_symbol", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("modifier", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("summary", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("currency", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("market", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("country", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("category_group", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("family", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("three_year_earnings_growth", sa.Float(), nullable=True),
        sa.Column("annual_holdings_turnover", sa.Float(), nullable=True),
        sa.Column("annual_report_expense_ratio", sa.Float(), nullable=True),
        sa.Column("holding_percent", sa.JSON(), nullable=True),
        sa.Column("median_market_cap", sa.Float(), nullable=True),
        sa.Column("price_to_book", sa.Float(), nullable=True),
        sa.Column("price_to_cashflow", sa.Float(), nullable=True),
        sa.Column("price_to_earnings", sa.Float(), nullable=True),
        sa.Column("price_to_sales", sa.Float(), nullable=True),
        sa.Column("sector_weightings", sa.JSON(), nullable=True),
        sa.Column("total_net_assets", sa.Float(), nullable=True),
        sa.Column("beta_3_year", sa.Float(), nullable=True),
        sa.Column("category", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("fund_family", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("fund_inception_date", sa.Float(), nullable=True),
        sa.Column("legal_type", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column(
            "long_business_summary", sqlmodel.sql.sqltypes.AutoString(), nullable=True
        ),
        sa.Column("long_name", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("nav_price", sa.Float(), nullable=True),
        sa.Column("quote_type", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("short_name", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("three_year_average_return", sa.Float(), nullable=True),
        sa.Column("trailing_annual_dividend_rate", sa.Float(), nullable=True),
        sa.Column("trailing_annual_dividend_yield", sa.Float(), nullable=True),
        sa.Column("trailing_pe", sa.Float(), nullable=True),
        sa.Column("yield_", sa.Float(), nullable=True),
        sa.Column("ytd_return", sa.Float(), nullable=True),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sqlite_autoincrement=True,
    )
    with op.batch_alter_table("etf", schema=None) as batch_op:
        batch_op.create_index(batch_op.f("ix_etf_source"), ["source"], unique=False)
        batch_op.create_index(batch_op.f("ix_etf_symbol"), ["symbol"], unique=False)

    op.create_table(
        "financialmetrics",
        sa.Column("symbol", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("exchange", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("source", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("date", sa.DateTime(), nullable=False),
        sa.Column("created_at", sa.Date(), nullable=False),
        sa.Column("ebitda", sa.Float(), nullable=True),
        sa.Column("net_income", sa.Float(), nullable=True),
        sa.Column("pe_ratio", sa.Float(), nullable=True),
        sa.Column("market_capitalization", sa.Float(), nullable=True),
        sa.Column("basic_eps", sa.Float(), nullable=True),
        sa.Column("diluted_eps", sa.Float(), nullable=True),
        sa.Column("total_revenue", sa.Float(), nullable=True),
        sa.Column("operating_revenue", sa.Float(), nullable=True),
        sa.Column("gross_profit", sa.Float(), nullable=True),
        sa.Column("total_expenses", sa.Float(), nullable=True),
        sa.Column("operating_income", sa.Float(), nullable=True),
        sa.Column("profit_margin", sa.Float(), nullable=True),
        sa.Column("cost_of_revenue", sa.Float(), nullable=True),
        sa.Column("tax_provision", sa.Float(), nullable=True),
        sa.Column("tax_rate", sa.Float(), nullable=True),
        sa.PrimaryKeyConstraint("symbol", "source", "date"),
    )
    with op.batch_alter_table("financialmetrics", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_financialmetrics_date"), ["date"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_financialmetrics_source"), ["source"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_financialmetrics_symbol"), ["symbol"], unique=False
        )

    op.create_table(
        "financialstracker",
        sa.Column("exchange", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("source", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("symbol", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("date", sa.Date(), nullable=False),
        sa.PrimaryKeyConstraint("source", "symbol"),
        sqlite_autoincrement=True,
    )
    with op.batch_alter_table("financialstracker", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_financialstracker_source"), ["source"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_financialstracker_symbol"), ["symbol"], unique=False
        )

    op.create_table(
        "price",
        sa.Column("exchange", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("symbol", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("source", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("date", sa.DateTime(), nullable=False),
        sa.Column("created_at", sa.Date(), nullable=False),
        sa.Column("open", sa.Float(), nullable=False),
        sa.Column("high", sa.Float(), nullable=False),
        sa.Column("low", sa.Float(), nullable=False),
        sa.Column("close", sa.Float(), nullable=False),
        sa.Column("volume", sa.Float(), nullable=False),
        sa.Column("dividends", sa.Float(), nullable=True),
        sa.Column("stock_splits", sa.Float(), nullable=True),
        sa.PrimaryKeyConstraint("symbol", "source", "date"),
    )
    with op.batch_alter_table("price", schema=None) as batch_op:
        batch_op.create_index(batch_op.f("ix_price_date"), ["date"], unique=False)
        batch_op.create_index(batch_op.f("ix_price_source"), ["source"], unique=False)
        batch_op.create_index(batch_op.f("ix_price_symbol"), ["symbol"], unique=False)

    op.create_table(
        "pricetracker",
        sa.Column("exchange", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("source", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("symbol", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("date", sa.Date(), nullable=False),
        sa.PrimaryKeyConstraint("source", "symbol"),
        sqlite_autoincrement=True,
    )
    with op.batch_alter_table("pricetracker", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_pricetracker_source"), ["source"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_pricetracker_symbol"), ["symbol"], unique=False
        )

    op.create_table(
        "quarterlybalancesheet",
        sa.Column("symbol", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("exchange", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("source", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("date", sa.DateTime(), nullable=False),
        sa.Column("created_at", sa.Date(), nullable=False),
        sa.Column("treasury_stock", sa.Float(), nullable=True),
        sa.Column("common_stock_shares_outstanding", sa.Float(), nullable=True),
        sa.Column("common_stock", sa.Float(), nullable=True),
        sa.Column("short_long_term_debt_total", sa.Float(), nullable=True),
        sa.Column("capital_lease_obligations", sa.Float(), nullable=True),
        sa.Column("total_shareholder_equity", sa.Float(), nullable=True),
        sa.Column("retained_earnings", sa.Float(), nullable=True),
        sa.Column("total_liabilities", sa.Float(), nullable=True),
        sa.Column("total_non_current_liabilities", sa.Float(), nullable=True),
        sa.Column("other_non_current_liabilities", sa.Float(), nullable=True),
        sa.Column("long_term_debt", sa.Float(), nullable=True),
        sa.Column("total_current_liabilities", sa.Float(), nullable=True),
        sa.Column("other_current_liabilities", sa.Float(), nullable=True),
        sa.Column("current_debt", sa.Float(), nullable=True),
        sa.Column("current_accounts_payable", sa.Float(), nullable=True),
        sa.Column("total_assets", sa.Float(), nullable=True),
        sa.Column("total_non_current_assets", sa.Float(), nullable=True),
        sa.Column("other_non_current_assets", sa.Float(), nullable=True),
        sa.Column(
            "accumulated_depreciation_amortization_ppe", sa.Float(), nullable=True
        ),
        sa.Column("property_plant_equipment", sa.Float(), nullable=True),
        sa.Column("total_current_assets", sa.Float(), nullable=True),
        sa.Column("other_current_assets", sa.Float(), nullable=True),
        sa.Column("inventory", sa.Float(), nullable=True),
        sa.Column("current_net_receivables", sa.Float(), nullable=True),
        sa.Column("cash_and_short_term_investments", sa.Float(), nullable=True),
        sa.Column(
            "cash_and_cash_equivalents_at_carrying_value", sa.Float(), nullable=True
        ),
        sa.PrimaryKeyConstraint("symbol", "source", "date"),
    )
    with op.batch_alter_table("quarterlybalancesheet", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_quarterlybalancesheet_date"), ["date"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_quarterlybalancesheet_source"), ["source"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_quarterlybalancesheet_symbol"), ["symbol"], unique=False
        )

    op.create_table(
        "quarterlycashflow",
        sa.Column("symbol", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("exchange", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("source", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("date", sa.DateTime(), nullable=False),
        sa.Column("created_at", sa.Date(), nullable=False),
        sa.Column("operating_cash_flow", sa.Float(), nullable=True),
        sa.Column("change_in_operating_liabilities", sa.Float(), nullable=True),
        sa.Column("change_in_working_capital", sa.Float(), nullable=True),
        sa.Column("change_in_other_working_capital", sa.Float(), nullable=True),
        sa.Column("change_in_receivables", sa.Float(), nullable=True),
        sa.Column("change_in_inventory", sa.Float(), nullable=True),
        sa.Column("depreciation_amortization_depletion", sa.Float(), nullable=True),
        sa.Column("capital_expenditure", sa.Float(), nullable=True),
        sa.Column("cash_flow_from_investing_activities", sa.Float(), nullable=True),
        sa.Column("financing_cash_flow", sa.Float(), nullable=True),
        sa.Column("repurchase_of_capital_stock", sa.Float(), nullable=True),
        sa.Column("cash_dividends_paid", sa.Float(), nullable=True),
        sa.Column("common_stock_dividend_paid", sa.Float(), nullable=True),
        sa.Column("proceeds_from_issuance_of_common_stock", sa.Float(), nullable=True),
        sa.Column("changes_in_cash", sa.Float(), nullable=True),
        sa.Column("net_income_from_continuing_operations", sa.Float(), nullable=True),
        sa.PrimaryKeyConstraint("symbol", "source", "date"),
    )
    with op.batch_alter_table("quarterlycashflow", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_quarterlycashflow_date"), ["date"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_quarterlycashflow_source"), ["source"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_quarterlycashflow_symbol"), ["symbol"], unique=False
        )

    op.create_table(
        "quarterlyfinancialmetrics",
        sa.Column("symbol", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("exchange", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("source", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("date", sa.DateTime(), nullable=False),
        sa.Column("created_at", sa.Date(), nullable=False),
        sa.Column("ebitda", sa.Float(), nullable=True),
        sa.Column("net_income", sa.Float(), nullable=True),
        sa.Column("pe_ratio", sa.Float(), nullable=True),
        sa.Column("market_capitalization", sa.Float(), nullable=True),
        sa.Column("basic_eps", sa.Float(), nullable=True),
        sa.Column("diluted_eps", sa.Float(), nullable=True),
        sa.Column("total_revenue", sa.Float(), nullable=True),
        sa.Column("operating_revenue", sa.Float(), nullable=True),
        sa.Column("gross_profit", sa.Float(), nullable=True),
        sa.Column("total_expenses", sa.Float(), nullable=True),
        sa.Column("operating_income", sa.Float(), nullable=True),
        sa.Column("profit_margin", sa.Float(), nullable=True),
        sa.Column("cost_of_revenue", sa.Float(), nullable=True),
        sa.Column("tax_provision", sa.Float(), nullable=True),
        sa.Column("tax_rate", sa.Float(), nullable=True),
        sa.PrimaryKeyConstraint("symbol", "source", "date"),
    )
    with op.batch_alter_table("quarterlyfinancialmetrics", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_quarterlyfinancialmetrics_date"), ["date"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_quarterlyfinancialmetrics_source"), ["source"], unique=False
        )
        batch_op.create_index(
            batch_op.f("ix_quarterlyfinancialmetrics_symbol"), ["symbol"], unique=False
        )

    op.create_table(
        "sources",
        sa.Column("source", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.PrimaryKeyConstraint("source"),
    )
    with op.batch_alter_table("sources", schema=None) as batch_op:
        batch_op.create_index(batch_op.f("ix_sources_source"), ["source"], unique=False)

    op.create_table(
        "view",
        sa.Column("symbol", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("exchange", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("source", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("date", sa.Date(), nullable=False),
        sa.Column("created_at", sa.Date(), nullable=False),
        sa.Column("name", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("isin", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("view_name", sqlmodel.sql.sqltypes.AutoString(), nullable=True),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sqlite_autoincrement=True,
    )
    with op.batch_alter_table("view", schema=None) as batch_op:
        batch_op.create_index(batch_op.f("ix_view_source"), ["source"], unique=False)
        batch_op.create_index(batch_op.f("ix_view_symbol"), ["symbol"], unique=False)

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("view", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_view_symbol"))
        batch_op.drop_index(batch_op.f("ix_view_source"))

    op.drop_table("view")
    with op.batch_alter_table("sources", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_sources_source"))

    op.drop_table("sources")
    with op.batch_alter_table("quarterlyfinancialmetrics", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_quarterlyfinancialmetrics_symbol"))
        batch_op.drop_index(batch_op.f("ix_quarterlyfinancialmetrics_source"))
        batch_op.drop_index(batch_op.f("ix_quarterlyfinancialmetrics_date"))

    op.drop_table("quarterlyfinancialmetrics")
    with op.batch_alter_table("quarterlycashflow", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_quarterlycashflow_symbol"))
        batch_op.drop_index(batch_op.f("ix_quarterlycashflow_source"))
        batch_op.drop_index(batch_op.f("ix_quarterlycashflow_date"))

    op.drop_table("quarterlycashflow")
    with op.batch_alter_table("quarterlybalancesheet", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_quarterlybalancesheet_symbol"))
        batch_op.drop_index(batch_op.f("ix_quarterlybalancesheet_source"))
        batch_op.drop_index(batch_op.f("ix_quarterlybalancesheet_date"))

    op.drop_table("quarterlybalancesheet")
    with op.batch_alter_table("pricetracker", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_pricetracker_symbol"))
        batch_op.drop_index(batch_op.f("ix_pricetracker_source"))

    op.drop_table("pricetracker")
    with op.batch_alter_table("price", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_price_symbol"))
        batch_op.drop_index(batch_op.f("ix_price_source"))
        batch_op.drop_index(batch_op.f("ix_price_date"))

    op.drop_table("price")
    with op.batch_alter_table("financialstracker", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_financialstracker_symbol"))
        batch_op.drop_index(batch_op.f("ix_financialstracker_source"))

    op.drop_table("financialstracker")
    with op.batch_alter_table("financialmetrics", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_financialmetrics_symbol"))
        batch_op.drop_index(batch_op.f("ix_financialmetrics_source"))
        batch_op.drop_index(batch_op.f("ix_financialmetrics_date"))

    op.drop_table("financialmetrics")
    with op.batch_alter_table("etf", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_etf_symbol"))
        batch_op.drop_index(batch_op.f("ix_etf_source"))

    op.drop_table("etf")
    with op.batch_alter_table("equity", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_equity_symbol"))
        batch_op.drop_index(batch_op.f("ix_equity_source"))
        batch_op.drop_index(batch_op.f("ix_equity_country"))

    op.drop_table("equity")
    with op.batch_alter_table("earningsdate", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_earningsdate_symbol"))
        batch_op.drop_index(batch_op.f("ix_earningsdate_source"))
        batch_op.drop_index(batch_op.f("ix_earningsdate_date"))

    op.drop_table("earningsdate")
    with op.batch_alter_table("currency", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_currency_symbol"))
        batch_op.drop_index(batch_op.f("ix_currency_source"))
        batch_op.drop_index(batch_op.f("ix_currency_base_currency"))

    op.drop_table("currency")
    with op.batch_alter_table("crypto", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_crypto_symbol"))
        batch_op.drop_index(batch_op.f("ix_crypto_source"))
        batch_op.drop_index(batch_op.f("ix_crypto_cryptocurrency"))

    op.drop_table("crypto")
    with op.batch_alter_table("cashflow", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_cashflow_symbol"))
        batch_op.drop_index(batch_op.f("ix_cashflow_source"))
        batch_op.drop_index(batch_op.f("ix_cashflow_date"))

    op.drop_table("cashflow")
    with op.batch_alter_table("balancesheet", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_balancesheet_symbol"))
        batch_op.drop_index(batch_op.f("ix_balancesheet_source"))
        batch_op.drop_index(batch_op.f("ix_balancesheet_date"))

    op.drop_table("balancesheet")
    with op.batch_alter_table("analysis", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_analysis_symbol"))
        batch_op.drop_index(batch_op.f("ix_analysis_source"))

    op.drop_table("analysis")
    # ### end Alembic commands ###



from typing import Literal

TickerOnlySources = Literal["investpy", "FMPAssets", "FinanceDatabase"]

Sources = Literal[
    "Tiingo",
    "investpy",
    "Yfinance",
    "FMP",
    "FMPAssets",
    "FinanceDatabase",
    "AlphaVantage",
    "YahooQuery",
]

SeriesLength = Literal["max", "1d", "5d", "1mo", "3mo", "6mo"]
DELAY = 0.2
